# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Hyperbolic stretch tool."""

from .stretch import StretchTool
from ..gtk.customwidgets import HBox, VBox, CheckButton, SpinButton
from ...imageprocessing import imageprocessing
from ...imageprocessing.stretchfunctions import ghyperbolic_stretch_function

class GeneralizedHyperbolicStretchTool(StretchTool):
  """Generalized hyperbolic stretch tool class, derived from the StretchTool class."""

  _action_ = "Stretching histograms (generalized hyperbolic stretch function)..."

  # Build window.

  _window_name_ = "Generalized hyperbolic stretch"

  def options_widgets(self, widgets):
    """Return a box with tool options widgets and store the reference to these widgets in container 'widgets'.
       Return None if there are no tool options widgets."""
    hbox = HBox()
    widgets.bindbutton = CheckButton(label = "Bind RGB channels")
    widgets.bindbutton.set_active(True)
    widgets.bindbutton.connect("toggled", lambda button: self.update("bindrgb"))
    hbox.pack(widgets.bindbutton, expand = True, fill = True)
    widgets.inversebutton = CheckButton(label = "Inverse transformation")
    widgets.inversebutton.connect("toggled", lambda button: self.update("inverse"))
    hbox.pack(widgets.inversebutton)
    return hbox

  def tab_widgets(self, key, widgets):
    """Return a box with tab widgets for channel 'key' in "R" (red), "G" (green), "B" (blue), "S" (saturation), "V" (value) or "L" (luma),
       and store the reference to these widgets in container 'widgets'.
       Return None if there is no tab for this channel."""
    if not key in ["R", "G", "B", "V", "L"]: return None
    percentiles = self.reference.stats["L"].percentiles
    step = (percentiles[2]-percentiles[0])/10. if percentiles is not None else .01
    step = min(max(step, .0001), .01)
    cbox = VBox(margin = 16)
    hbox = HBox()
    cbox.pack(hbox)
    widgets.logD1spin = SpinButton(0., 0., 10., .1, digits = 3)
    widgets.logD1spin.connect("value-changed", lambda button: self.update("D"))
    hbox.pack("Global log(D+1):")
    hbox.pack(widgets.logD1spin)
    widgets.Bspin = SpinButton(0, -5., 15., .1, digits = 3)
    widgets.Bspin.connect("value-changed", lambda button: self.update("B"))
    hbox.pack(5*" "+"Local B:")
    hbox.pack(widgets.Bspin)
    widgets.SYPspin = SpinButton(.5, 0., 1., step/2., digits = 5)
    widgets.SYPspin.connect("value-changed", lambda button: self.update("SYP"))
    hbox.pack(5*" "+"Symmetry point:")
    hbox.pack(widgets.SYPspin)
    hbox = HBox()
    cbox.pack(hbox)
    widgets.SPPspin = SpinButton(0., 0., .99, step/2., digits = 5)
    widgets.SPPspin.connect("value-changed", lambda button: self.update("SPP"))
    hbox.pack("Shadow protection point:")
    hbox.pack(widgets.SPPspin)
    widgets.HPPspin = SpinButton(1., .01, 1., step, digits = 5)
    widgets.HPPspin.connect("value-changed", lambda button: self.update("HPP"))
    hbox.pack(5*" "+"Highlight protection point:")
    hbox.pack(widgets.HPPspin)
    if key == "L":
      widgets.highlightsbutton = CheckButton(label = "Protect highlights")
      widgets.highlightsbutton.connect("toggled", lambda button: self.update(None))
      hbox.pack(5*" ")
      hbox.pack(widgets.highlightsbutton)
    return cbox

  # Tool methods.

  def get_params(self):
    """Return tool parameters."""
    params = {}
    for key in self.channelkeys:
      channel = self.widgets.channels[key]
      logD1 = channel.logD1spin.get_value()
      B = channel.Bspin.get_value()
      SYP = channel.SYPspin.get_value()
      SPP = channel.SPPspin.get_value()
      HPP = channel.HPPspin.get_value()
      params[key] = (logD1, B, SYP, SPP, HPP)
    params["highlights"] = self.widgets.channels["L"].highlightsbutton.get_active()
    params["inverse"] = self.widgets.inversebutton.get_active()
    params["rgbluma"] = imageprocessing.get_rgb_luma()
    return params

  def set_params(self, params):
    """Set tool parameters 'params'."""
    unbindrgb = False
    redparams = params["R"]
    for key in self.channelkeys:
      channel = self.widgets.channels[key]
      if key in ("R", "G", "B"):
        unbindrgb = unbindrgb or (params[key] != redparams)
      logD1, B, SYP, SPP, HPP = params[key]
      channel.logD1spin.set_value_block(logD1)
      channel.Bspin.set_value_block(B)
      channel.SYPspin.set_value_block(SYP)
      channel.SPPspin.set_value_block(SPP)
      channel.HPPspin.set_value_block(HPP)
    self.widgets.channels["L"].highlightsbutton.set_active_block(params["highlights"])
    self.widgets.inversebutton.set_active(params["inverse"])
    if unbindrgb: self.widgets.bindbutton.set_active_block(False)
    self.update("all")

  def run(self, params):
    """Run tool for parameters 'params'."""
    self.image.copy_image_from(self.reference)
    transformed = False
    inverse = params["inverse"]
    for key in self.channelkeys:
      logD1, B, SYP, SPP, HPP = params[key]
      outofrange = self.outofrange and key in ["R", "G", "B"]
      if not outofrange and logD1 == 0.: continue
      transformed = True
      self.image.generalized_stretch(ghyperbolic_stretch_function, (logD1, B, SYP, SPP, HPP, inverse), channels = key)
    if transformed and params["highlights"]: self.image.protect_highlights()
    return params, transformed

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    operation = "GHStretch("
    if params["inverse"]: operation = "Inverse"+operation
    separator = ""
    for key in self.channelkeys:
      logD1, B, SYP, SPP, HPP = params[key]
      if key == "L":
        red, green, blue = params["rgbluma"]
        operation += f"{separator}L({red:.2f}, {green:.2f}, {blue:.2f}) : (log(D+1) = {logD1:.3f}, B = {B:.3f}, SYP = {SYP:.5f}, SPP = {SPP:.5f}, HPP = {HPP:.5f})"
      else:
        operation += f"{separator}{key} : (log(D+1) = {logD1:.3f}, B = {B:.3f}, SYP = {SYP:.5f}, SPP = {SPP:.5f}, HPP = {HPP:.5f})"
      separator = ", "
    if params["highlights"]: operation += ", protect highlights"
    operation += ")"
    return operation

  # Plot histograms, stretch function, display stats...

  def stretch_function(self, t, params):
    """Return the stretch function f(t) for parameters 'params'."""
    return ghyperbolic_stretch_function(t, params)

  def add_histogram_widgets(self, ax):
    """Add histogram widgets (other than stretch function) in axes 'ax'."""
    self.widgets.SPPline = ax.axvline(0., linestyle = "-.", zorder = -2)
    self.widgets.SYPline = ax.axvline(.5, linestyle = "-.", zorder = -2)
    self.widgets.HPPline = ax.axvline(1., linestyle = "-.", zorder = -2)

  # Update histograms, stats... on widget or key_press events.

  def update_widgets(self, key, changed):
    """Update widgets (other than histograms and stats) on change of 'changed' in channel 'key'."""
    channel = self.widgets.channels[key]
    logD1 = channel.logD1spin.get_value()
    B = channel.Bspin.get_value()
    SYP = channel.SYPspin.get_value()
    SPP = channel.SPPspin.get_value()
    HPP = channel.HPPspin.get_value()
    if changed == "SPP":
      if SPP > HPP-.005:
        SPP = HPP-.005
        channel.SPPspin.set_value_block(SPP)
      if SPP > SYP:
        SYP = SPP
        channel.SYPspin.set_value_block(SYP)
    elif changed == "HPP":
      if HPP < SPP+.005:
        HPP = SPP+.005
        channel.HPPspin.set_value_block(HPP)
      if HPP < SYP:
        SYP = HPP
        channel.SYPspin.set_value_block(SYP)
    elif changed == "SYP":
      if SYP < SPP:
        SPP = SYP
        channel.SPPspin.set_value_block(SPP)
      elif SYP > HPP:
        HPP = SYP
        channel.HPPspin.set_value_block(HPP)
    color = channel.color
    lcolor = channel.lcolor
    self.widgets.SPPline.set_xdata([SPP, SPP])
    self.widgets.SPPline.set_color(.1*lcolor)
    self.widgets.SYPline.set_xdata([SYP, SYP])
    self.widgets.SYPline.set_color(.5*lcolor)
    self.widgets.HPPline.set_xdata([HPP, HPP])
    self.widgets.HPPline.set_color(.9*lcolor)
    inverse = self.widgets.inversebutton.get_active()
    self.plot_stretch_function(lambda t: self.stretch_function(t, (logD1, B, SYP, SPP, HPP, inverse)), color)
    if self.widgets.bindbutton.get_active() and key in ("R", "G", "B"):
      for rgbkey in ("R", "G", "B"):
        rgbchannel = self.widgets.channels[rgbkey]
        rgbchannel.logD1spin.set_value_block(logD1)
        rgbchannel.Bspin.set_value_block(B)
        rgbchannel.SYPspin.set_value_block(SYP)
        rgbchannel.SPPspin.set_value_block(SPP)
        rgbchannel.HPPspin.set_value_block(HPP)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Luma RGB dialog."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from .gtk.customwidgets import HBox, VBox, HButtonBox, Button, SpinButton

class LumaRGBDialog(Gtk.Window):
  """Luma RGB dialog class."""

  # Here we store the data, widgets & methods of this simple dialog directly in the window object.

  def __init__(self, parent, callback, rgbluma):
    """Open a luma RGB dialog for parent window 'parent', with initial
       RGB components 'rgbluma'. When the apply button is pressed, close the
       dialog and call 'callback(rgbluma)', where rgbluma are the updated RGB
       components of the luma."""
    super().__init__(title = "Luma RGB",
                     transient_for = parent,
                     modal = True,
                     border_width = 16)
    wbox = VBox()
    self.add(wbox)
    hbox = HBox()
    wbox.pack(hbox)
    self.redspin = SpinButton(rgbluma[0], 0., 1., .01, digits = 2)
    hbox.pack("Red:")
    hbox.pack(self.redspin)
    self.greenspin = SpinButton(rgbluma[1], 0., 1., .01, digits = 2)
    hbox.pack(8*" "+"Green:")
    hbox.pack(self.greenspin)
    self.bluespin = SpinButton(rgbluma[2], 0., 1., .01, digits = 2)
    hbox.pack(8*" "+"Blue:")
    hbox.pack(self.bluespin)
    hbox = HButtonBox()
    wbox.pack(hbox)
    applybutton = Button(label = "OK")
    applybutton.connect("clicked", self.apply)
    hbox.pack(applybutton)
    humanbutton = Button(label = "Human")
    humanbutton.connect("clicked", self.set_human_vision)
    hbox.pack(humanbutton)
    uniformbutton = Button(label = "Uniform")
    uniformbutton.connect("clicked", self.set_uniform_rgb)
    hbox.pack(uniformbutton)
    cancelbutton = Button(label = "Cancel")
    cancelbutton.connect("clicked", lambda button: self.destroy())
    hbox.pack(cancelbutton)
    self.callback = callback
    self.show_all()

  def apply(self, *args, **kwargs):
    """Apply luma RGB settings."""
    red = self.redspin.get_value()
    green = self.greenspin.get_value()
    blue = self.bluespin.get_value()
    total = red+green+blue
    if total <= 0.: return
    rgbluma = (red/total, green/total, blue/total)
    self.callback(rgbluma)
    self.destroy()

  def set_human_vision(self, *args, **kwargs):
    """Set human vision luma RGB components."""
    self.redspin.set_value(.3)
    self.greenspin.set_value(.6)
    self.bluespin.set_value(.1)

  def set_uniform_rgb(self, *args, **kwargs):
    """Set uniform luma RGB components."""
    self.redspin.set_value(1./3.)
    self.greenspin.set_value(1./3.)
    self.bluespin.set_value(1./3.)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Color noise reduction tool."""

from ..gtk.customwidgets import HBox, VBox, CheckButton, RadioButtons, HScale, ComboBoxText
from ..toolmanager import BaseToolWindow
from ...imageprocessing.utils import scale_pixels
from ...imageprocessing.colors import srgb_to_lrgb, lrgb_luminance, lrgb_to_srgb
import numpy as np

class ColorNoiseReductionTool(BaseToolWindow):
  """Color noise reduction tool class."""

  _action_ = "Reducing color noise..."

  _help_ = """Reduce color noise for the red (R), yellow (Y), green (G), cyan (C), blue (B) or magenta (M) hue. For the green hue for example,

    \u2022 G \u2192 min(G, m) with m = (R+B)/2 for average neutral protection.
    \u2022 G \u2192 min(G, m) with m = max(R, B) for maximum neutral protection.
    \u2022 G \u2192 G[(1-mixing)+m*mixing] with m = (R+B)/2 for additive mask protection.
    \u2022 G \u2192 G[(1-mixing)+m*mixing] with m = max(R, B) for maximum mask protection.

The RGB components of each pixel are rescaled to preserve the CIE lightness L* if the "preserve lightness" checkbox is ticked."""

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Color noise reduction"): return False
    self.reference.luminance = self.reference.srgb_luminance() # Compute reference luminance.
    #self.reference.lightscale = lrgb_luminance(np.clip(self.reference.rgb, 0., 1.)**2.2)**(1./2.2) # Approximate back and forth transformation between sRGB & lRGB color spaces.
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.colorbuttons = RadioButtons(("None", "None"), ("red", "Red"), ("yellow", "Yellow"), ("green", "Green"), \
                                             ("cyan", "Cyan"), ("blue", "Blue"), ("magenta", "Magenta"))
    self.widgets.colorbuttons.connect("toggled", lambda button: self.update("color"))
    wbox.pack(self.widgets.colorbuttons.hbox(prepend = "Color:"))
    hbox = HBox()
    wbox.pack(hbox)
    self.widgets.modelcombo = ComboBoxText(("AvgNeutral", "Average neutral protection"), ("MaxNeutral", "Maximal neutral protection"), \
                                           ("AddMask", "Additive mask protection"), ("MaxMask", "Maximum mask protection"))
    self.widgets.modelcombo.connect("changed", lambda combo: self.update("model"))
    hbox.pack("Model:")
    hbox.pack(self.widgets.modelcombo)
    self.widgets.lightnessbutton = CheckButton(label = "Preserve lightness")
    self.widgets.lightnessbutton.set_active(True)
    self.widgets.lightnessbutton.connect("toggled", lambda button: self.update("lightness"))
    hbox.pack(self.widgets.lightnessbutton)
    self.widgets.mixingscale = HScale(1., 0., 1., .01, digits = 2, marks = [0., 1.], length = 480, expand = False)
    self.widgets.mixingscale.set_sensitive(False)
    self.widgets.mixingscale.connect("value-changed", lambda scale: self.update("mixing"))
    wbox.pack(self.widgets.mixingscale.hbox(prepend = "Mixing:"))
    #self.widgets.thresholdscale = HScale(0., 0., 1., .01, digits = 2, marks = [0., 1.], length = 480, expand = False)
    #self.widgets.thresholdscale.connect("value-changed", lambda scale: self.update("threshold"))
    #wbox.pack(self.widgets.thresholdscale.hbox(prepend = "Threshold:"))
    wbox.pack(self.tool_control_buttons())
    self.start(identity = True)
    return True

  def get_params(self):
    """Return tool parameters."""
    color = self.widgets.colorbuttons.get_selected()
    model = self.widgets.modelcombo.get_selected()
    mixing = self.widgets.mixingscale.get_value()
    #threshold = self.widgets.thresholdscale.get_value()
    lightness = self.widgets.lightnessbutton.get_active()
    #return color, model, mixing, threshold, lightness
    return color, model, mixing, lightness

  def set_params(self, params):
    """Set tool parameters 'params'."""
    #color, model, mixing, threshold, lightness = params
    color, model, mixing, lightness = params
    self.widgets.colorbuttons.set_selected_block(color)
    self.widgets.modelcombo.set_selected_block(model)
    self.widgets.mixingscale.set_value_block(mixing)
    #self.widgets.thresholdscale.set_value_block(threshold)
    self.widgets.lightnessbutton.set_active_block(lightness)
    self.update("all")

  def run(self, params):
    """Run tool for parameters 'params'."""
    #color, model, mixing, threshold, lightness = params
    color, model, mixing, lightness = params
    self.image.copy_image_from(self.reference)
    if color == "red":
      icc, ic1, ic2, negative = 0, 1, 2, False
    elif color == "yellow":
      icc, ic1, ic2, negative = 2, 0, 1, True
    elif color == "green":
      icc, ic1, ic2, negative = 1, 0, 2, False
    elif color == "cyan":
      icc, ic1, ic2, negative = 0, 1, 2, True
    elif color == "blue":
      icc, ic1, ic2, negative = 2, 0, 1, False
    elif color == "magenta":
      icc, ic1, ic2, negative = 1, 0, 2, True
    else:
      return params, False
    self.image.clip() # Clip before reducing color noise.
    if negative: self.image.negative()
    rgb = self.image.rgb
    #mask = (rgb[icc] >= threshold)
    if model == "AvgNeutral":
      m = (rgb[ic1]+rgb[ic2])/2.
      #rgb[icc] = np.where(mask, np.minimum(rgb[icc], m, rgb[icc])
      rgb[icc] = np.minimum(rgb[icc], m)
    elif model == "MaxNeutral":
      m = np.maximum(rgb[ic1], rgb[ic2])
      #rgb[icc] = np.where(mask, np.minimum(rgb[icc], m, rgb[icc])
      rgb[icc] = np.minimum(rgb[icc], m)
    elif model == "AddMask":
      m = np.minimum(1., rgb[ic1]+rgb[ic2])
      #rgb[icc] *= np.where(mask, (1.-mixing)+m*mixing, 1.)
      rgb[icc] *= (1.-mixing)+m*mixing
    else:
      m = np.maximum(rgb[ic1], rgb[ic2])
      #rgb[icc] *= np.where(mask, (1.-mixing)+m*mixing, 1.)
      rgb[icc] *= (1.-mixing)+m*mixing
    if negative: self.image.negative()
    if lightness:
      lrgb = srgb_to_lrgb(self.image.rgb)
      self.image.rgb = lrgb_to_srgb(scale_pixels(lrgb, lrgb_luminance(lrgb), self.reference.luminance))
      #lightscale = lrgb_luminance(self.image.rgb**2.2)**(1./2.2)     # Approximate back and forth transformation between sRGB & lRGB color spaces.
      #self.image.scale_pixels(lightscale, self.reference.lightscale) # This preserves the sRGB hue, yet not exactly the lightness.
      difflight = self.image.srgb_lightness()-self.reference.srgb_lightness()
      print(f"Maximum lightness difference = {abs(difflight).max()}.")
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    #color, model, mixing, threshold, lightness = params
    color, model, mixing, lightness = params
    if color == "None": return None
    operation = f"ReduceColorNoise(color = {color}, model = {model}"
    if model in ["AddMask", "MaxMask"]:
      operation += f", mixing = {mixing:.2f}"
    #operation += f", threshold = {threshold:.2f}"
    if lightness: operation += f", preserve L*"
    operation += ")"
    return operation

  # Update widgets.

  def update(self, changed):
    """Update widgets on change of 'changed'."""
    if changed in ["model", "all"]:
      model = self.widgets.modelcombo.get_selected()
      self.widgets.mixingscale.set_sensitive(model in ["AddMask", "MaxMask"])
    self.reset_polling(self.get_params()) # Expedite main window update.



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Misc Gtk utilities."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

# Shortcuts.
from gi.repository.GLib import markup_escape_text

class Container:
  """Empty class as a container."""
  pass

def get_work_area(window):
  """Return the width and height of the monitor displaying 'window'."""
  screen = window.get_screen()
  display = screen.get_display()
  monitor = display.get_monitor_at_window(screen.get_root_window())
  workarea = monitor.get_workarea()
  return workarea.width, workarea.height

def flush_gtk_events():
  """Flush all pending gtk events."""
  while Gtk.events_pending(): Gtk.main_iteration()




# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Image statistics window."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from .gtk.customwidgets import HBox, VBox, HButtonBox, Button
from .gtk.keyboard import decode_key
from .base import BaseWindow, FigureCanvas, BaseToolbar, Container
from .misc.utils import histogram_bins, plot_histograms, highlight_histogram
from matplotlib.figure import Figure

class StatsWindow(BaseWindow):
  """Image statistics window class."""

  key = None # Registered key of the image tab.

  def open(self, image, key, tab):
    """Open statistics window for image 'image' of tab 'tab' with key 'key'."""
    if self.opened: self.close()
    self.key = key
    self.opened = True
    self.window = Gtk.Window(title = f"Histograms & statistics of tab '{tab}'", transient_for = self.app.mainwindow.window, destroy_with_parent = True, border_width = 16)
    self.window.connect("delete-event", self.close)
    self.window.connect("key-press-event", self.key_press)
    self.widgets = Container()
    wbox = VBox()
    self.window.add(wbox)
    fbox = VBox(spacing = 0)
    wbox.pack(fbox, expand = True, fill = True)
    self.widgets.fig = Figure(figsize = (10., 3.), layout = "constrained")
    canvas = FigureCanvas(self.widgets.fig)
    canvas.set_size_request(600, 180)
    fbox.pack(canvas, expand = True, fill = True)
    toolbar = BaseToolbar(canvas, self.widgets.fig)
    fbox.pack(toolbar)
    wbox.pack("Press [L] to toggle lin/log scale")
    stats = image.statistics()
    self.widgets.selection = self.pack_image_statistics_treeview(stats, wbox)
    hbox = HButtonBox()
    wbox.pack(hbox)
    self.widgets.closebutton = Button(label = "Close")
    self.widgets.closebutton.connect("clicked", self.close)
    hbox.pack(self.widgets.closebutton)
    self.widgets.fig.histax = self.widgets.fig.add_subplot(111)
    self.histlogscale = False
    self.histcolors = ((1., 0., 0.), (0., 1., 0.), (0., 0., 1.), (0., 0., 0.), (.5, 0.5, 0.5))
    self.histograms = image.histograms(nbins = histogram_bins(stats["L"], self.app.get_color_depth()))
    self.plot_histograms()
    self.widgets.selection.connect("changed", lambda selection: self.highlight_histogram())
    self.window.show_all()

  def close(self, *args, **kwargs):
    """Close statistics window.
       If the kwarg 'key' is not None, close only if self.key = key."""
    if not self.opened: return
    key = kwargs.get("key", None)
    if key is not None and self.key != key: return
    self.window.destroy()
    self.opened = False
    self.key = None
    del self.widgets
    del self.histograms

  def pack_image_statistics_treeview(self, stats, box):
    """Pack image statistics 'stats' (see imageprocessing.Image.statistics) as a TreeView in box 'box'.
       Return a TreeView selection object to get the selected channel with self.get_selected_channel()."""
    width, height, npixels = stats["L"].width, stats["L"].height, stats["L"].npixels
    box.pack(f"Image size = {width}x{height} pixels = {npixels} pixels")
    store = Gtk.ListStore(int, str, str, str, str, str, str, str, str, str, str)
    idx = 0
    for key in ("R", "G", "B", "V", "L"):
      channel = stats[key]
      if channel.median is None:
        perc25 = "-"
        median = "-"
        perc75 = "-"
      else:
        perc25 = f"{channel.percentiles[0]:.5f}"
        median = f"{channel.median:.5f}"
        perc75 = f"{channel.percentiles[2]:.5f}"
      store.append([idx, channel.name, f"{channel.minimum:.5f}", perc25, median, perc75, f"{channel.maximum:.5f}",
                    f"{channel.zerocount:d}", f"({100.*channel.zerocount/npixels:6.3f}%)", f"{channel.outcount:d}", f"({100.*channel.outcount/npixels:6.3f}%)"])
      idx += 1
    treeview = Gtk.TreeView(model = store, search_column = -1)
    box.pack(treeview)
    renderer = Gtk.CellRendererText()
    renderer.set_property("xalign", 0.)
    column = Gtk.TreeViewColumn("Channel", renderer, text = 1)
    column.set_alignment(0.)
    column.set_expand(True)
    treeview.append_column(column)
    renderer = Gtk.CellRendererText()
    renderer.set_property("xalign", 1.)
    column = Gtk.TreeViewColumn("Minimum", renderer, text = 2)
    column.set_alignment(1.)
    column.set_expand(True)
    treeview.append_column(column)
    renderer = Gtk.CellRendererText()
    renderer.set_property("xalign", 1.)
    column = Gtk.TreeViewColumn("25th %", renderer, text = 3)
    column.set_alignment(1.)
    column.set_expand(True)
    treeview.append_column(column)
    renderer = Gtk.CellRendererText()
    renderer.set_property("xalign", 1.)
    column = Gtk.TreeViewColumn("median", renderer, text = 4)
    column.set_alignment(1.)
    column.set_expand(True)
    treeview.append_column(column)
    renderer = Gtk.CellRendererText()
    renderer.set_property("xalign", 1.)
    column = Gtk.TreeViewColumn("75th %", renderer, text = 5)
    column.set_alignment(1.)
    column.set_expand(True)
    treeview.append_column(column)
    renderer = Gtk.CellRendererText()
    renderer.set_property("xalign", 1.)
    column = Gtk.TreeViewColumn("Maximum", renderer, text = 6)
    column.set_alignment(1.)
    column.set_expand(True)
    treeview.append_column(column)
    column = Gtk.TreeViewColumn("Zeros (<= 0)")
    column.set_alignment(1.)
    column.set_expand(True)
    count = Gtk.CellRendererText()
    count.set_property("xalign", 1.)
    column.pack_start(count, True)
    column.add_attribute(count, "text", 7)
    percent = Gtk.CellRendererText()
    percent.set_property("xalign", 1.)
    column.pack_start(percent, False)
    column.add_attribute(percent, "text", 8)
    treeview.append_column(column)
    column = Gtk.TreeViewColumn("Out-of-range (> 1)")
    column.set_alignment(1.)
    column.set_expand(True)
    count = Gtk.CellRendererText()
    count.set_property("xalign", 1.)
    column.pack_start(count, True)
    column.add_attribute(count, "text", 9)
    percent = Gtk.CellRendererText()
    percent.set_property("xalign", 1.)
    column.pack_start(percent, False)
    column.add_attribute(percent, "text", 10)
    treeview.append_column(column)
    box.pack("The medians and percentiles (%) above exclude pixels &lt;= 0 and &gt;= 1")
    selection = treeview.get_selection()
    selection.set_mode(Gtk.SelectionMode.SINGLE)
    return selection

  def get_selected_channel(self):
    """Return selected channel (-1 if None)."""
    model, list_iter = self.widgets.selection.get_selected()
    return model[list_iter][0] if list_iter is not None else -1

  def plot_histograms(self):
    """Plot histograms."""
    ax = self.widgets.fig.histax
    ax.histlines = plot_histograms(ax, *self.histograms, colors = self.histcolors, ylogscale = self.histlogscale)
    highlight_histogram(ax.histlines, self.get_selected_channel())

  def highlight_histogram(self):
    """Highlight histogram line."""
    highlight_histogram(self.widgets.fig.histax.histlines, self.get_selected_channel())
    self.widgets.fig.canvas.draw_idle()

  def key_press(self, widget, event):
    """Callback for key press in the statistics window."""
    key = decode_key(event)
    if key.ctrl or key.alt: return
    if key.uname == "L": # Toggle log scale.
      self.histlogscale = not self.histlogscale
      self.plot_histograms()
      self.widgets.fig.canvas.draw_idle()
      self.window.queue_draw()
      return True



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Keyboard management."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk
from .utils import Container

def decode_key(event):
  """Decode keypress event and return a container 'key':
       - key.ctrl is True if "Ctrl" is pressed.
       - key.alt  is Trye if "Alt"  is pressed.
       - key.name is the key name.
       - key.uname is the key name in upper case letters."""
  key = Container()
  key.ctrl = event.state & Gdk.ModifierType.CONTROL_MASK
  key.alt  = event.state & Gdk.ModifierType.MOD1_MASK
  key.name = Gdk.keyval_name(event.keyval)
  key.uname = key.name.upper()
  return key



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Custom Gtk widgets."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, GObject
from .signals import Signals
from collections import OrderedDict as OD

# Shortcuts.
from gi.repository.Gtk import Align

###########
# Labels. #
###########

class Label(Gtk.Label):
  """Gtk label with markup enabled."""

  def __init__(self, label = "", markup = True, **kwargs):
    """Initialize a Gtk.Label with label 'label' and default halign = Gtk.Align.START.
       Markup is enabled unless 'markup' is False. All other kwargs are passed to Gtk.Label."""
    kwargs.setdefault("halign", Align.START)
    Gtk.Label.__init__(self, **kwargs)
    if markup:
      self.set_markup(label)
    else:
      self.set_label(label)

##########
# Boxes. #
##########

class HBox(Gtk.HBox):
  """Gtk horizontal box with default settings & wrappers."""

  def __init__(self, *args, **kwargs):
    """Initialize a Gtk.HBox with default spacing = 8."""
    kwargs.setdefault("spacing", 8)
    Gtk.HBox.__init__(self, *args, **kwargs)

  def pack(self, widget, expand = False, fill = False, padding = 0):
    """Wrapper for Gtk.HBox.pack_start(widget, expand, fill, padding) with default expand = False, fill = False and padding = 0.
       If a string, 'widget' is converted into a Gtk label."""
    if isinstance(widget, str): widget = Label(widget)
    self.pack_start(widget, expand, fill, padding)

#

class VBox(Gtk.VBox):
  """Gtk vertical box with default settings & wrappers."""

  def __init__(self, *args, **kwargs):
    """Initialize a Gtk.VBox with default spacing = 16."""
    kwargs.setdefault("spacing", 16)
    Gtk.VBox.__init__(self, *args, **kwargs)

  def pack(self, widget, expand = False, fill = False, padding = 0):
    """Wrapper for Gtk.VBox.pack_start(widget, expand, fill, padding) with default expand = False, fill = False and padding = 0.
       If a string, 'widget' is converted into a Gtk label."""
    if isinstance(widget, str): widget = Label(widget)
    self.pack_start(widget, expand, fill, padding)

#

class FramedHBox():
  """Framed horizontal box with default settings & wrappers."""

  def __new__(cls, label = None, align = (.05, .5), **kwargs):
    """Initialize a framed HBox with default margin = 16 (with respect to the frame).
       'label' is the label of the frame. The position of the label within the frame is controlled
       by 'align' (see Gtk.Frame.set_label_align). All other kwargs are passed to HBox.
       Returns the Gtk frame widget and the HBox."""
    kwargs.setdefault("margin", 16)
    hbox = HBox(**kwargs)
    frame = Gtk.Frame(label = label)
    frame.set_label_align(*align)
    frame.add(hbox)
    return frame, hbox

#

class FramedVBox():
  """Framed vertical box with default settings & wrappers."""

  def __new__(cls, label = None, align = (.05, .5), **kwargs):
    """Initialize a framed VBox with default margin = 16 (with respect to the frame).
       'label' is the label of the frame. The position of the label within the frame is controlled
       by 'align' (see Gtk.Frame.set_label_align). All other kwargs are passed to VBox.
       Returns the Gtk frame widget and the HBox."""
    kwargs.setdefault("margin", 16)
    vbox = VBox(**kwargs)
    frame = Gtk.Frame(label = label)
    frame.set_label_align(*align)
    frame.add(vbox)
    return frame, vbox

#

class HButtonBox(Gtk.HButtonBox):
  """Gtk horizontal button box with default settings & wrappers."""

  def __init__(self, *args, **kwargs):
    """Initialize a Gtk.HButtonBox with default homogeneous = True, spacing = 16 and halign = Gtk.Align.START."""
    kwargs.setdefault("homogeneous", True)
    kwargs.setdefault("spacing", 16)
    kwargs.setdefault("halign", Align.START)
    Gtk.HButtonBox.__init__(self, *args, **kwargs)

  def pack(self, widget, expand = False, fill = False, padding = 0):
    """Wrapper for Gtk.HButtonBox.pack_start(widget, expand, fill, padding) with default expand = False, fill = False and padding = 0."""
    self.pack_start(widget, expand, fill, padding)

#

class ScrolledBox(Gtk.ScrolledWindow):
  """Gtk scrolled window with default settings & wrappers."""

  def __init__(self, width, height, *args, **kwargs):
    """Initialize a Gtk.ScrolledWindow with minimal width 'width' and minimal height 'height'."""
    Gtk.ScrolledWindow.__init__(self, *args, **kwargs)
    self.set_min_content_width(width)
    self.set_min_content_height(height)

#

def pack_hbox(widget, prepend = None, append = None, spacing = 8, expand = False):
  """Return a HBox with widgets 'prepend', 'widget', and 'append' spaced by 'spacing'.
     If strings, 'prepend' and 'append' are converted into labels.
     The 'widget' is packed as expandable & filling if 'expand' is True."""
  hbox = HBox(spacing = spacing)
  if prepend is not None:
    if isinstance(prepend, str): prepend = Label(prepend)
    hbox.pack(prepend)
  hbox.pack(widget, expand = expand, fill = expand)
  if append is not None:
    if isinstance(append, str): append = Label(append)
    hbox.pack(append)
  return hbox

##########
# Grids. #
##########

class Grid(Gtk.Grid):
  """Gtk grid with default settings & wrappers."""

  def __init__(self, *args, **kwargs):
    """Initialize a Gtk.Grid with default column_spacing = 8."""
    kwargs.setdefault("column_spacing", 8)
    Gtk.Grid.__init__(self, *args, **kwargs)

  def attach(self, widget, left, top, width = 1, height = 1):
    """Wrapper for Gtk.Grid.attach(widget, left, top, width, height) with default width = 1 and height = 1.
       If a string, 'widget' is converted into a Gtk label."""
    if isinstance(widget, str): widget = Label(widget)
    Gtk.Grid.attach(self, widget, left, top, width, height)

############
# Buttons. #
############

class Button(Signals, Gtk.Button):
  """Custom Gtk button with extended signal management."""

  def __init__(self, *args, **kwargs):
    """Initialize class."""
    Signals.__init__(self)
    Gtk.Button.__init__(self, *args, **kwargs)

#

class HoldButton(Signals, Gtk.Button):
  """Custom Gtk "hold" button with extended signal management.
     When pressed, this button emits a "hold" signal every 'delay' ms,
     then a "clicked" signal once released."""

  __gsignals__ = {"hold": (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE, ())}

  def __init__(self, *args, delay = 500, **kwargs):
    """Initialize class. 'delay' is the button hold delay (default 500 ms)."""
    Signals.__init__(self)
    Gtk.Button.__init__(self, *args, **kwargs)
    self.delay = delay
    self.connect("pressed", self.__pressed)
    self.connect("released", self.__released)

  def __pressed(self, widget):
    """Internal callback for button press events."""
    self.timer = GObject.timeout_add(self.delay, self.__longpressed)

  def __longpressed(self):
    """Internal callback for button long press events."""
    self.emit("hold")
    return True

  def __released(self, widget):
    """Internal callback for button release events."""
    GObject.source_remove(self.timer)

#

class CheckButton(Signals, Gtk.CheckButton):
  """Custom Gtk check button with extended signal management."""

  def __init__(self, *args, **kwargs):
    """Initialize class."""
    Signals.__init__(self)
    Gtk.CheckButton.__init__(self, *args, **kwargs)

  def set_active_block(self, *args, **kwargs):
    """Set button status, blocking all signals (no callbacks)."""
    self.block_all_signals()
    self.set_active(*args, **kwargs)
    self.unblock_all_signals()

#

class RadioButton(Signals, Gtk.RadioButton):
  """Custom Gtk radio button with extended signal management."""

  def __init__(self, *args, **kwargs):
    """Initialize class."""
    Signals.__init__(self)
    Gtk.RadioButton.__init__(self, *args, **kwargs)

  @classmethod
  def new_with_label_from_widget(cls, widget, label):
    """Add a radio button with label 'label' to the group of widget 'widget'."""
    button = cls(label = label)
    if widget is not None: button.join_group(widget)
    return button

  def set_active_block(self, *args, **kwargs):
    """Set button status, blocking all signals (no callbacks)."""
    self.block_all_signals()
    self.set_active(*args, **kwargs)
    self.unblock_all_signals()

#

class RadioButtons:
  """Group of custom Gtk radio buttons."""

  def __init__(self, *args):
    """Initialize a group of custom Gtk radio buttons defined by *args, a
       list of tuples (key, label), where 'key' is a key that uniquely identifies
       the button and 'label' is the button label."""
    firstbutton = None
    self.buttons = OD()
    for key, label in args:
      if key in self.buttons.keys():
        raise KeyError(f"The key '{key}' is already registered.")
      button = RadioButton.new_with_label_from_widget(firstbutton, label)
      button.key = key
      self.buttons[key] = button
      if firstbutton is None: firstbutton = button

  def get_selected(self):
    """Return the key of the selected radio button."""
    for key, button in self.buttons.items():
      if button.get_active(): return key
    return None # Shall not happen !

  def set_selected(self, key):
    """Select the radio button with key 'key'."""
    try:
      self.buttons[key].set_active(True)
    except KeyError:
      raise KeyError(f"There is no button with key '{key}'.")

  def set_selected_block(self, key):
    """Select the radio button with key 'key', blocking all signals (no callbacks)."""
    try:
      self.buttons[key].set_active_block(True)
    except KeyError:
      raise KeyError(f"There is no button with key '{key}'.")

  def connect(self, *args, **kwargs):
    """Connect signal & callback to all buttons."""
    for button in self.buttons.values():
      button.connect(*args, **kwargs)

  def hbox(self, prepend = None, append = None, spacing = 8):
    """Return a HBox with widget 'prepend', the radio buttons, and widget 'append', spaced by 'spacing'.
       If strings, 'prepend' and 'append' are converted into labels."""
    buttonsbox = HBox(spacing = spacing)
    for button in self.buttons.values():
      buttonsbox.pack(button)
    return pack_hbox(buttonsbox, prepend, append, spacing, False)

##########################
# Spin buttons & scales. #
##########################

class SpinButton(Signals, Gtk.SpinButton):
  """Custom Gtk spin button with extended signal management."""

  def __init__(self, value, minimum, maximum, step, page = None, digits = 2, climbrate = .01):
    """Return a Gtk spin button with current value 'value', minimum value 'minimum', maximum value 'maximum',
       step size 'step', page size 'page' (10*step if None), number of displayed digits 'digits', and climb rate 'climbrate'."""
    Signals.__init__(self)
    Gtk.SpinButton.__init__(self)
    self.configure(Gtk.Adjustment(value = value, lower = minimum, upper = maximum,
                                  step_increment = step, page_increment = 10*step if page is None else page), climbrate, digits)
    self.set_numeric(True)
    self.set_update_policy(Gtk.SpinButtonUpdatePolicy.IF_VALID)

  def set_value_block(self, *args, **kwargs):
    """Set value, blocking all signals (no callbacks)."""
    self.block_all_signals()
    self.set_value(*args, **kwargs)
    self.unblock_all_signals()

  def hbox(self, prepend = None, append = None, spacing = 8):
    """Return a HBox with widget 'prepend', the spin button, and widget 'append', spaced by 'spacing'.
       If strings, 'prepend' and 'append' are converted into labels."""
    return pack_hbox(self, prepend, append, spacing, False)

#

class HScale(Signals, Gtk.Scale):
  """Custom Gtk horizontal scale with extended signal management."""

  def __init__(self, value, minimum, maximum, step, page = None, marks = None, digits = 2, length = -1, expand = True):
    """Return a horizontal Gtk scale with current value 'value', minimum value 'minimum', maximum value 'maximum',
       step size 'step', page size 'page' (10*step if None), marks 'marks', number of displayed digits 'digits',
       and default length 'length' expandable if 'expand' is True."""
    Signals.__init__(self)
    Gtk.Scale.__init__(self)
    self.set_adjustment(Gtk.Adjustment(value = value, lower = minimum, upper = maximum,
                                       step_increment = step, page_increment = 10*step if page is None else page))
    self.set_orientation(Gtk.Orientation.HORIZONTAL)
    if marks is not None:
      for mark in marks:
        self.add_mark(mark, Gtk.PositionType.BOTTOM, f"{mark:.{digits}f}")
      self.set_value_pos(Gtk.PositionType.TOP)
    else:
      self.set_value_pos(Gtk.PositionType.RIGHT)
    self.set_value(value)
    self.set_draw_value(True)
    self.set_digits(digits)
    self.set_size_request(length, -1)
    self.set_hexpand(expand)
    self.expand = expand

  def set_value_block(self, *args, **kwargs):
    """Set value, blocking all signals (no callbacks)."""
    self.block_all_signals()
    self.set_value(*args, **kwargs)
    self.unblock_all_signals()

  def hbox(self, prepend = None, append = None, spacing = 8):
    """Return a HBox with widget 'prepend', the horizontal scale, and widget 'append', spaced by 'spacing'.
       If strings, 'prepend' and 'append' are converted into labels."""
    return pack_hbox(self, prepend, append, spacing, self.expand)

#

class HScaleSpinButton():
  """Custom Gtk horizontal scale coupled to a custom Gtk spin button, with extended signal management."""

  def __init__(self, value, minimum, maximum, step, page = None, digits = 2, length = -1, expand = True, climbrate = .01):
    """Return a Gtk scale/spin button with current value 'value', minimum value 'minimum', maximum value 'maximum',
       step size 'step', page size 'page' (10*step if None), number of displayed digits 'digits', default length 'length'
       expandable if 'expand' is True, and climb rate 'climbrate'."""
    self.expand = expand
    self.scale = HScale(value, minimum, maximum, step, page = page, digits = digits, length = length, expand = expand)
    self.button = SpinButton(value, minimum, maximum, step, page = page, digits = digits, climbrate = climbrate)
    self.scale.set_draw_value(False)
    self.scale.connect("value-changed", lambda scale: self.__sync(self.button, self.scale))
    self.button.connect("value-changed", lambda button: self.__sync(self.scale, self.button))
    self.__callbacks__ = {}

  def __sync(self, target, source):
    """Synchronize the value of widget 'target' with the value of widget 'source'."""
    target.set_value_block(source.get_value())
    callback = self.__callbacks__.get("value-changed", None)
    if callback is not None: return callback(self)

  def connect(self, signal, callback):
    """Connect signal 'signal' to callback(self)."""
    if signal not in ["value-changed"]:
      raise ValueError("Unknown signal.")
    self.__callbacks__[signal] = callback

  def get_value(self):
    """Return Gtk scale/spin button value."""
    return self.button.get_value()

  def set_value(self, value):
    """Set Gtk scale/spin button value 'value'."""
    self.button.set_value(value)

  def set_value_block(self, value):
    """Set Gtk scale/spin button value 'value', blocking all signals (no callbacks)."""
    self.scale.set_value_block(value)
    self.button.set_value_block(value)

  def layout1(self, label = None, spacing = 8):
    """Single line layout.
       Return a HBox with label 'label' (if not None), the Gtk scale and spin button, spaced by 'spacing'."""
    hbox = HBox(spacing = spacing)
    if label is not None: hbox.pack(Label(label))
    hbox.pack(self.scale, expand = self.expand, fill = self.expand)
    hbox.pack(self.button)
    return hbox

  def layout2(self, label = "", spacing = 8):
    """Two lines layout.
       Return a VBox with label 'label' and the Gtk spin button on one line (spaced by 'spacing'),
       and the Gtk scale on an other."""
    vbox = VBox(spacing = 0)
    hbox = HBox(spacing = spacing)
    vbox.pack(hbox)
    hbox.pack(Label(label), expand = True, fill = True)
    hbox.pack(self.button)
    vbox.pack(self.scale, expand = self.expand, fill = self.expand)
    return vbox

################
# Combo boxes. #
################

class ComboBoxText(Signals, Gtk.ComboBoxText):
  """Custom Gtk combo box with extended signal management."""

  def __init__(self, *args):
    """Initialize a Gtk combo box with items defined by *args, a list of tuples (key, label),
       where 'key' is a key that uniquely identifies the item and 'label' is the item label."""
    Signals.__init__(self)
    Gtk.ComboBoxText.__init__(self)
    self.keys = []
    for key, label in args:
      if key in self.keys:
        raise KeyError(f"The key '{key}' is already registered.")
      self.keys.append(key)
      self.append_text(label)
    self.set_active(0)

  def get_selected(self):
    """Return the key of the selected combo box item."""
    return self.keys[self.get_active()]

  def set_selected(self, key):
    """Select the combo box item with key 'key'."""
    try:
      self.set_active(self.keys.index(key))
    except ValueError:
      raise KeyError(f"There is no item with key '{key}'.")

  def set_selected_block(self, key):
    """Select the combo box item with key 'key', blocking all signals (no callbacks)."""
    self.block_all_signals()
    try:
      self.set_active(self.keys.index(key))
    except ValueError:
      raise KeyError(f"There is no item with key '{key}'.")
    self.unblock_all_signals()

  def hbox(self, prepend = None, append = None, spacing = 8):
    """Return a HBox with widget 'prepend', the combo box, and widget 'append', spaced by 'spacing'.
       If strings, 'prepend' and 'append' are converted into labels."""
    return pack_hbox(self, prepend, append, spacing, False)

###############
# Text input. #
###############

class Entry(Signals, Gtk.Entry):
  """Custom Gtk entry with extended signal management."""

  def __init__(self, text = "", width = -1):
    """Return a Gtk entry with default text 'text' and width 'width' (in chars)."""
    Signals.__init__(self)
    Gtk.Entry.__init__(self)
    self.expand = (width <= 0)
    self.set_width_chars(width)
    self.set_text(text)

  def set_text_block(self, *args, **kwargs):
    """Set text, blocking all signals (no callbacks)."""
    self.block_all_signals()
    self.set_text(*args, **kwargs)
    self.unblock_all_signals()

  def hbox(self, prepend = None, append = None, spacing = 8):
    """Return a HBox with widget 'prepend', the entry, and widget 'append', spaced by 'spacing'.
       If strings, 'prepend' and 'append' are converted into labels."""
    return pack_hbox(self, prepend, append, spacing, self.expand)

#

class TextBuffer(Signals, Gtk.TextBuffer):
  """Custom Gtk text buffer with extended signal management."""

  def __init__(self):
    """Return a Gtk text buffer."""
    Signals.__init__(self)
    Gtk.TextBuffer.__init__(self)

  def set_text_block(self, *args, **kwargs):
    """Set text, blocking all signals (no callbacks)."""
    self.block_all_signals()
    self.set_text(*args, **kwargs)
    self.unblock_all_signals()

  def append_text(self, text):
    """Append text 'text'."""
    self.insert(self.get_end_iter(), text, -1)

  def append_text_block(self, text):
    """Append text 'text', blocking all signals."""
    self.block_all_signals()
    self.insert(self.get_end_iter(), text, -1)
    self.unblock_all_signals()

  def append_markup(self, text):
    """Append markup text 'text'."""
    self.insert_markup(self.get_end_iter(), text, -1)

  def append_markup_block(self, text):
    """Append markup text 'text', blocking all signals."""
    self.block_all_signals()
    self.insert_markup(self.get_end_iter(), text, -1)
    self.unblock_all_signals()

  def copy_to_clipboard(self, *args, **kwargs):
    """Copy the content of the text buffer to the clipboard."""
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    clipboard.set_text(self.get_text(self.get_start_iter(), self.get_end_iter(), False), -1)

#

class TextView(Signals, Gtk.TextView):
  """Custom Gtk text view with extended signal management."""

  def __init__(self, textbuffer = None, editable = False, wrap = True, justification = Gtk.Justification.LEFT):
    """Return a Gtk text view with text buffer 'textbuffer' (created if None).
       The text view is editable if 'editable' is True, the text is wrapped if 'wrapped' is True, and is justified
       according to 'justification'."""
    Signals.__init__(self)
    Gtk.TextView.__init__(self)
    self.set_editable(editable)
    self.set_cursor_visible(editable)
    self.set_wrap_mode(wrap)
    self.set_justification(justification)
    if textbuffer is None: textbuffer = TextBuffer()
    self.set_buffer(textbuffer)

  def get_text(self, *args, **kwargs):
    """Get text."""
    return self.get_buffer().get_text()

  def set_text(self, *args, **kwargs):
    """Set text."""
    self.get_buffer().set_text(*args, **kwargs)

  def set_text_block(self, *args, **kwargs):
    """Set text, blocking all signals (no callbacks)."""
    self.get_buffer().set_text_block(*args, **kwargs)

  def append_text(self, text):
    """Append text 'text'."""
    self.get_buffer().append_text(text)

  def append_text_block(self, text):
    """Append text 'text', blocking all signals."""
    self.get_buffer().append_text_block(text)

  def append_markup(self, text):
    """Append markup text 'text'."""
    self.get_buffer().append_markup(text)

  def append_markup_block(self, text):
    """Append markup text 'text', blocking all signals."""
    self.get_buffer().append_markup_block(text)

  def copy_to_clipboard(self, *args, **kwargs):
    """Copy the content of the text view to the clipboard."""
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    self.get_buffer().copy_to_clipboard(*args, **kwargs)

##############
# Notebooks. #
##############

class Notebook(Signals, Gtk.Notebook):
  """Custom Gtk notebook with extended signal management."""

  def __init__(self, *args, pos = Gtk.PositionType.TOP, **kwargs):
    """Initialize Gtk notebook with tabs at position 'pos'."""
    Signals.__init__(self)
    Gtk.Notebook.__init__(self, *args, **kwargs)
    self.set_tab_pos(pos)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Resampling tool."""

from ..gtk.customwidgets import HBox, VBox, Label, SpinButton, HScaleSpinButton
from ..base import ErrorDialog
from ..toolmanager import BaseToolWindow
from ...imageprocessing import imageprocessing

class ResampleTool(BaseToolWindow):
  """Resampling tool."""

  _action_ = "Resampling image..."

  _onthefly_ = False # This transformation can not be applied on the fly.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Resampling image"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.rwidth, self.rheight = self.reference.size()
    self.aratio = self.rwidth/self.rheight
    wbox.pack(Label(f"Original image size: {self.rwidth}x{self.rheight} pixels"))
    hbox = HBox()
    wbox.pack(hbox)
    hbox.pack(Label("Target image size:"))
    self.widgets.widthbutton = SpinButton(self.rwidth, self.rwidth//4, 4*self.rwidth, 1., digits = 0)
    self.widgets.widthbutton.connect("value-changed", lambda button: self.update("width"))
    hbox.pack(self.widgets.widthbutton)
    hbox.pack(Label("x"))
    self.widgets.heightbutton = SpinButton(self.rheight, self.rheight//4, 4*self.rheight, 1., digits = 0)
    self.widgets.heightbutton.connect("value-changed", lambda button: self.update("height"))
    hbox.pack(self.widgets.heightbutton)
    hbox.pack(Label("pixels"))
    self.widgets.scalescale = HScaleSpinButton(1., .25, 4., .1, digits = 3, length = 320)
    self.widgets.scalescale.connect("value-changed", lambda button: self.update("scale"))
    wbox.pack(self.widgets.scalescale.layout2(label = "Scale:"))
    wbox.pack(self.tool_control_buttons(model = "applyonce"))
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return int(round(self.widgets.widthbutton.get_value())), int(round(self.widgets.heightbutton.get_value()))

  def set_params(self, params):
    """Set tool parameters 'params'."""
    width, height = params
    self.widgets.widthbutton.set_value(width)

  def run(self, params):
    """Run tool for parameters 'params'."""

    def Error(message):
      """Open error message dialog."""
      ErrorDialog(self.window, message)
      return False

    width, height = params
    try:
      self.image = self.reference.resize(width, height, inplace = False)
    except Exception as err:
      self.queue_gui_mainloop(Error, str(err))
      return (self.rwidth, self.rheight), False
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    width, height = params
    return f"Resample(size = {width}x{height} pixels)"

 # Update widgets.

  def update(self, changed):
    """Update widgets."""
    if changed == "width":
      width = round(self.widgets.widthbutton.get_value())
      height = round(width/self.aratio)
      scale = width/self.rwidth
    elif changed == "height":
      height = round(self.widgets.heightbutton.get_value())
      width = round(height*self.aratio)
      scale = width/self.rwidth
    else:
      scale = self.widgets.scalescale.get_value()
      width = round(self.rwidth*scale)
      height = round(self.rheight*scale)
    self.widgets.widthbutton.set_value_block(width)
    self.widgets.heightbutton.set_value_block(height)
    self.widgets.scalescale.set_value_block(scale)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Base window classes and widgets."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from .gtk.utils import Container
from matplotlib.backends.backend_gtk3agg import FigureCanvasGTK3Agg as FigureCanvas
from matplotlib.backends.backend_gtk3 import NavigationToolbar2GTK3 as NavigationToolbar

class BaseWindow:
  """Generic application window."""

  def __init__(self, app):
    """Bind the window with application 'app'."""
    self.app = app
    self.opened = False

class InfoDialog(Gtk.MessageDialog):
  """Modal info dialog class."""

  def __init__(self, parent, message):
    """Open a modal info dialog showing 'message' for window 'parent'."""
    super().__init__(title = "Information",
                     transient_for = parent,
                     destroy_with_parent = True,
                     message_type = Gtk.MessageType.INFO,
                     buttons = Gtk.ButtonsType.OK,
                     modal = True)
    self.set_markup(message)
    self.run()
    self.destroy()

class ErrorDialog(Gtk.MessageDialog):
  """Modal error dialog class."""

  def __init__(self, parent, message):
    """Open a modal error dialog showing 'message' for window 'parent'."""
    super().__init__(title = "Error",
                     transient_for = parent,
                     destroy_with_parent = True,
                     message_type = Gtk.MessageType.ERROR,
                     buttons = Gtk.ButtonsType.CLOSE,
                     modal = True)
    self.set_markup(message)
    self.run()
    self.destroy()

class BaseToolbar(NavigationToolbar):
  """A custom matplotlib navigation toolbar for integration in Gtk.
     This is the default matplotlib navigation toolbar with the
     "Forward", "Back", "Subplots", and "Save" buttons disabled."""

  def __init__(self, canvas, parent):
    toolitems = []
    previous = None
    for tool in self.toolitems:
      name = tool[0]
      if name is None:
        keep = previous is not None
      else:
        keep = name not in ("Forward", "Back", "Subplots", "Save")
      if keep:
        toolitems.append(tool)
        previous = name
    if toolitems:
      if toolitems[-1][0] is None: toolitems.pop(-1)
    self.toolitems = tuple(toolitems)
    try:
      super().__init__(canvas, parent) # Old API.
    except:
      super().__init__(canvas) # New API.



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Main window."""

import os
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, GObject
from .gtk.utils import get_work_area
from .gtk.customwidgets import Align, Label, HBox, VBox, Button, CheckButton, HScale, Notebook
from .gtk.keyboard import decode_key
from . import menus
from .base import BaseWindow, FigureCanvas, BaseToolbar, Container
from .luma import LumaRGBDialog
from .statistics import StatsWindow
from .lightcurve import LightCurveWindow
from ..imageprocessing import imageprocessing
import numpy as np
from matplotlib.figure import Figure
from collections import OrderedDict as OD

class MainWindow:
  """Main window class."""

  MAXIMGSIZE = .75 # Maximal width/height of the image (as a fraction of the screen resolution).

  SHADOWCOLOR = np.array([[1.], [.5], [0.]], dtype = imageprocessing.IMGTYPE)
  HIGHLIGHTCOLOR = np.array([[1.], [1.], [0.]], dtype = imageprocessing.IMGTYPE)
  DIFFCOLOR = np.array([[1.], [1.], [0.]], dtype = imageprocessing.IMGTYPE)

  _HELP_ = """[PAGE DOWN]: Next image tab
[PAGE UP]: Previous image tab
[D] : Show image description (if available)
[S]: Statistics (of the zoomed area)
[CTRL+C]: Copy image in a new tab
[CTRL+V]: Paste tab parameters to the tool
[CTRL+X]: Close image tab (if possible)
[CTRL+TAB]: Toggle between main and tool windows""" # Help tootip.

  def __init__(self, app):
    """Bind the window with application 'app'."""
    self.app = app

  def open(self):
    """Open main window."""
    self.window = Gtk.ApplicationWindow(application = self.app, title = "eQuimageGtk v"+self.app.version)
    self.window.connect("delete-event", self.close)
    self.window.connect("key-press-event", self.key_press)
    self.window.connect("key-release-event", self.key_release)
    self.widgets = Container()
    wbox = VBox(spacing = 0)
    self.window.add(wbox)
    fig = Figure()
    ax = fig.add_axes([0., 0., 1., 1.])
    self.canvas = FigureCanvas(fig)
    self.canvas.size = (-1, -1)
    wbox.pack(self.canvas, expand = True, fill = True)
    hbox = HBox(spacing = 0)
    wbox.pack(hbox)
    self.tabs = Notebook(pos = Gtk.PositionType.BOTTOM)
    self.tabs.set_scrollable(True)
    self.tabs.set_show_border(False)
    self.tabs.connect("switch-page", lambda tabs, tab, itab: self.display_tab(itab))
    hbox.pack(self.tabs, expand = True, fill = True)
    label = Label("[?]")
    label.set_tooltip_text(self._HELP_)
    hbox.pack(label, padding = 8)
    hbox = HBox(spacing = 0)
    wbox.pack(hbox)
    hbox.pack("Output range Min:", padding = 4)
    self.widgets.minscale = HScale(0., 0., 1., .01, digits = 2, length = 128)
    self.widgets.minscale.connect("value-changed", lambda scale: self.update_output_range("Min"))
    hbox.pack(self.widgets.minscale, expand = True, fill = True, padding = 4)
    self.widgets.spinner = Gtk.Spinner()
    hbox.pack(self.widgets.spinner, padding = 4)
    hbox.pack("Max:", padding = 4)
    self.widgets.maxscale = HScale(1., 0., 1., .01, digits = 2, length = 128)
    self.widgets.maxscale.connect("value-changed", lambda scale: self.update_output_range("Max"))
    hbox.pack(self.widgets.maxscale, expand = True, fill = True, padding = 4)
    hbox = HBox(spacing = 0)
    wbox.pack(hbox)
    self.widgets.redbutton = CheckButton(label = "Red")
    self.widgets.redbutton.set_active(True)
    self.widgets.redbutton.connect("toggled", lambda button: self.update_channels("R"))
    hbox.pack(self.widgets.redbutton)
    self.widgets.greenbutton = CheckButton(label = "Green")
    self.widgets.greenbutton.set_active(True)
    self.widgets.greenbutton.connect("toggled", lambda button: self.update_channels("G"))
    hbox.pack(self.widgets.greenbutton)
    self.widgets.bluebutton = CheckButton(label = "Blue")
    self.widgets.bluebutton.set_active(True)
    self.widgets.bluebutton.connect("toggled", lambda button: self.update_channels("B"))
    hbox.pack(self.widgets.bluebutton)
    self.widgets.lumabutton = CheckButton(label = self.rgb_luma_string())
    self.widgets.lumabutton.set_active(False)
    self.widgets.lumabutton.connect("toggled", lambda button: self.update_channels("L"))
    hbox.pack(self.widgets.lumabutton)
    self.widgets.rgblumabutton = Button(label = "Set", halign = Align.START)
    self.widgets.rgblumabutton.connect("clicked", lambda button: LumaRGBDialog(self.window, self.set_rgb_luma, self.get_rgb_luma()))
    hbox.pack(self.widgets.rgblumabutton, expand = True, fill = True)
    self.widgets.shadowbutton = CheckButton(label = "Shadowed")
    self.widgets.shadowbutton.set_active(False)
    self.widgets.shadowbutton.connect("toggled", lambda button: self.update_modifiers("S"))
    hbox.pack(self.widgets.shadowbutton)
    self.widgets.highlightbutton = CheckButton(label = "Highlighted")
    self.widgets.highlightbutton.set_active(False)
    self.widgets.highlightbutton.connect("toggled", lambda button: self.update_modifiers("H"))
    hbox.pack(self.widgets.highlightbutton)
    self.widgets.diffbutton = CheckButton(label = "Differences")
    self.widgets.diffbutton.set_active(False)
    self.widgets.diffbutton.connect("toggled", lambda button: self.update_modifiers("D"))
    hbox.pack(self.widgets.diffbutton)
    class MainToolbar(BaseToolbar): home = self.home # Work around matplotlib toolbar limitations.
    self.widgets.toolbar = MainToolbar(self.canvas, fig)
    wbox.pack(self.widgets.toolbar)
    self.set_copy_paste_callbacks(None, None)
    self.set_rgb_luma_callback(None)
    self.set_guide_lines(None)
    self.popup = None
    # Add context menu for statistics & light curve to the canvas.
    self.statswindow = StatsWindow(self.app)
    self.lightwindow = LightCurveWindow(self.app)
    self.contextmenu = Gtk.Menu().new_from_model(menus.builder.get_object("MainWindowContextMenu"))
    self.contextmenu.attach_to_widget(self.window)
    self.canvas.connect("button-press-event", self.button_press)
    self.reset_images()
    self.window.show_all()

  def close(self, *args, **kwargs):
    """Quit application."""
    dialog = Gtk.MessageDialog(transient_for = self.window,
                               message_type = Gtk.MessageType.QUESTION,
                               buttons = Gtk.ButtonsType.OK_CANCEL,
                               modal = True)
    dialog.set_markup("Are you sure you want to quit ?")
    response = dialog.run()
    dialog.destroy()
    if response != Gtk.ResponseType.OK: return True
    print("Exiting eQuimageGtk...")
    self.app.quit()

  # Images/tabs associations.

  def get_current_tab(self):
    """Return current tab."""
    return self.tabs.get_current_page()

  def set_current_tab(self, tab):
    """Set current tab 'tab'."""
    if self.get_current_tab() != tab: self.tabs.set_current_page(tab)

  def display_tab(self, tab):
    """Display image of tab 'tab'."""
    keys = list(self.images.keys())
    self.draw_image(keys[tab])

  def update_tab_label(self, tab, label):
    """Update the label 'label' of tab 'tab'."""
    self.tabs.set_tab_label(self.tabs.get_nth_page(tab), Label(label))

  def get_keys(self):
    """Return the list of image keys."""
    return list(self.images.keys())

  def get_key_tab(self, tab):
    """Return the image key of tab 'tab'."""
    return list(self.images.keys())[tab]

  def get_tab_key(self, key):
    """Return the tab of image 'key'."""
    try:
      return list(self.images.keys()).index(key)
    except KeyError:
      raise KeyError(f"There is no image with key '{key}'.")

  def get_current_key(self):
    """Return the image key of the current tab."""
    tab = self.get_current_tab()
    return self.get_key_tab(tab) if tab >= 0 else None

  def set_current_key(self, key):
    """Set current tab with image 'key'."""
    tab = self.get_tab_key(key)
    if tab is not None: self.set_current_tab(tab)

  def update_key_label(self, key, label):
    """Update the tab label 'label' of image 'key'."""
    tab = self.get_tab_key(key)
    if tab is not None: self.update_tab_label(tab, label)

  # Update displayed channels.

  def update_channels(self, toggled):
    """Update channels buttons."""
    if toggled == "L":
      luma = self.widgets.lumabutton.get_active()
      self.widgets.redbutton.set_active_block(not luma)
      self.widgets.greenbutton.set_active_block(not luma)
      self.widgets.bluebutton.set_active_block(not luma)
    else:
      red = self.widgets.redbutton.get_active()
      green = self.widgets.greenbutton.get_active()
      blue = self.widgets.bluebutton.get_active()
      self.widgets.lumabutton.set_active_block(not (red or green or blue))
    self.draw_image(self.get_current_key())

  # Update image modifiers (shadow, highlight, difference).

  def update_modifiers(self, toggled):
    """Update image modifiers."""
    if toggled == "D":
      if self.widgets.diffbutton.get_active():
        self.widgets.shadowbutton.set_active_block(False)
        self.widgets.highlightbutton.set_active_block(False)
    else:
      self.widgets.diffbutton.set_active_block(False)
    modifiers = self.widgets.shadowbutton.get_active() or self.widgets.highlightbutton.get_active() or self.widgets.diffbutton.get_active()
    self.widgets.minscale.set_sensitive(not modifiers)
    self.widgets.maxscale.set_sensitive(not modifiers)
    self.draw_image(self.get_current_key())

  # Update output range.

  def update_output_range(self, updated):
    """Update output range."""
    vmin = self.widgets.minscale.get_value()
    vmax = self.widgets.maxscale.get_value()
    if vmax-vmin < .01:
      if updated == "Max":
        vmin = max(0., vmax-.01)
        vmax = vmin+.01
      else:
        vmax = min(vmin+.01, 1.)
        vmin = vmax-.01
      self.widgets.minscale.set_value_block(vmin)
      self.widgets.maxscale.set_value_block(vmax)
    self.refresh_image()

  # Image modifiers (shadow, highlight, difference).

  def difference(self, image, reference, channels):
    """Highlight differences between 'image' and 'reference' with color DIFFCOLOR."""
    if reference is None: return
    if reference.shape != image.shape: return
    mask = np.any(image[channels] != reference[channels], axis = 0)
    image[:, mask] = self.DIFFCOLOR

  def shadow_highlight(self, image, reference, channels, shadow = True, highlight = True):
    """If shadow is True,
         show pixels black on 'image' and     on 'reference' with color .5*SHADOWCOLOR,
         and  pixels black on 'image' but not on 'reference' with color     SHADOWCOLOR,
       If higlight is True,
         show pixels with at least one channel >= 1 on 'image' and     on  'reference' with color .5*HIGHLIGHTCOLOR,
         and  pixels with at least one channel >= 1 on 'image' but not on  'reference' with color     HIGHLIGHTCOLOR."""
    if shadow:
      shadowmask = np.all(image[channels] < imageprocessing.IMGTOL, axis = 0)
    if highlight:
      hlightmask = np.any(image[channels] > 1.-imageprocessing.IMGTOL, axis = 0)
    if shadow:
      image[:, shadowmask] = self.SHADOWCOLOR
      if reference is not None:
        if reference.shape == image.shape:
          refmask = np.all(reference[channels] < imageprocessing.IMGTOL, axis = 0)
          image[:, shadowmask & refmask] = .5*self.SHADOWCOLOR
    if highlight:
      image[:, hlightmask] = self.HIGHLIGHTCOLOR
      if reference is not None:
        if reference.shape == image.shape:
          refmask = np.any(reference[channels] > 1.-imageprocessing.IMGTOL, axis = 0)
          image[:, hlightmask & refmask] = .5*self.HIGHLIGHTCOLOR

  # Draw or refresh the image displayed in the main window.

  def set_canvas_size(self, width, height):
    """Set canvas size for a figure width 'width' and height 'height'."""
    swidth, sheight = get_work_area(self.window)
    cwidth, cheight = self.MAXIMGSIZE*swidth, self.MAXIMGSIZE*swidth*height/width
    if cheight > self.MAXIMGSIZE*sheight:
      cwidth, cheight = self.MAXIMGSIZE*sheight*width/height, self.MAXIMGSIZE*sheight
    if self.canvas.size != (cwidth, cheight):
      self.canvas.size = (cwidth, cheight)
      self.canvas.set_size_request(cwidth, cheight)
      self.window.resize(1, 1)

  def draw_image(self, key):
    """Apply modifiers and draw image with key 'key'."""
    if key is None: return
    try:
      image = self.images[key]
    except KeyError:
      raise KeyError(f"There is no image with key '{key}'.")
    reference = None
    shadow = self.widgets.shadowbutton.get_active()
    highlight = self.widgets.highlightbutton.get_active()
    diff = self.widgets.diffbutton.get_active()
    modifiers = shadow or highlight or diff
    luma = self.widgets.lumabutton.get_active()
    if luma:
      image = np.repeat(image._luma_[np.newaxis], 3, axis = 0)
      channels = np.array([True, False, False])
      if modifiers and self.refkey is not None:
        reference = np.repeat(self.images[self.refkey]._luma_[np.newaxis], 3, axis = 0)
    else:
      image = image.get_image_copy()
      channels = np.array([self.widgets.redbutton.get_active(), self.widgets.greenbutton.get_active(), self.widgets.bluebutton.get_active()])
      image[~channels] = 0.
      if modifiers and self.refkey is not None:
        reference = self.images[self.refkey].get_image()
    if modifiers:
      if diff:
        self.difference(image, reference, channels)
      else:
        self.shadow_highlight(image, reference, channels, shadow, highlight)
    self.refresh_image(image)
    self.set_idle()

  def refresh_image(self, image = None, redraw = False):
    """Draw 'image' (if not None) or refresh the current image.
       Reset the axes and redraw the whole canvas if 'redraw' is True."""
    update = not redraw # Is this an update or fresh plot ?
    update = update and self.currentimage is not None
    if image is not None:
      currentshape = self.currentimage.shape if update else None
      self.currentimage = np.clip(np.moveaxis(image, 0, -1), 0., 1.)
      update = update and self.currentimage.shape == currentshape
    if self.currentimage is None: return # Nothing to draw !
    vmin = self.widgets.minscale.get_value() if self.widgets.minscale.get_sensitive() else 0.
    vmax = self.widgets.maxscale.get_value() if self.widgets.maxscale.get_sensitive() else 1.
    if vmin > 0. or vmax < 1.:
      ranged = np.where((self.currentimage >= vmin) & (self.currentimage <= vmax), self.currentimage, 0.)
    else:
      ranged = self.currentimage
    ax = self.canvas.figure.axes[0]
    if update:
      ax.imshown.set_data(ranged) # Update image (preserves zoom, ...).
    else:
      ax.clear() # Draw image.
      ax.imshown = ax.imshow(ranged, aspect = "equal")
      ax.axis("off")
      if self.plot_guide_lines is not None: self.plot_guide_lines(ax)
    self.canvas.draw_idle()
    self.window.queue_draw()

  def home(self, *args, **kwargs):
    """Replacement callback for matplotlib toolbar home button.
       Matplotlib toolbar home button resets the axes to their original span even if the image
       has been resampled and changed size in between. Therefore, resampled images appear cropped.
       This replacement callback handles image size changes properly by redrawing the canvas completely."""
    self.refresh_image(redraw = True)

  # Manage the dictionary of images displayed in the tabs.

  def reset_images(self):
    """Reset main window images."""
    self.images = None
    self.currentimage = None
    nimages = self.app.get_nbr_images()
    if nimages > 0:
      self.set_canvas_size(*self.app.get_image_size())
      if nimages > 3:
        self.set_images(OD(Image = self.app.get_image(-1), Original = self.app.get_image(1)), reference = "Original")
      elif nimages > 0:
        self.set_images(OD(Original = self.app.get_image(1)), reference = "Original")
    else:
      self.set_canvas_size(800, 600)
      try:
        splash = imageprocessing.load_image(os.path.join(self.app.get_packagepath(), "images", "splash.png"), {"tag": "Welcome"})
      except:
        splash = imageprocessing.black_image(800, 600, {"tag": "Welcome"})
      self.set_images(OD(Splash = splash))

  def set_images(self, images, reference = None):
    """Set main window images and reference."""
    self.close_key_windows()
    self.tabs.block_all_signals()
    for tab in range(self.tabs.get_n_pages()): self.tabs.remove_page(-1)
    self.images = OD()
    for key, image in images.items():
      self.images[key] = image.ref()
      self.images[key]._luma_ = self.images[key].luma()
    self.refkey = None
    if reference is not None:
      if reference not in self.images.keys():
        raise KeyError(f"There is no image with key '{reference}'.")
      self.refkey = reference
    for key, image in self.images.items():
      label = self.images[key].meta.get("tag", key)
      if key == self.refkey: label += " (\u2022)"
      self.tabs.append_page(Gtk.Alignment(), Label(label)) # Append a zero size dummy child.
    self.widgets.redbutton.set_active_block(True)
    self.widgets.greenbutton.set_active_block(True)
    self.widgets.bluebutton.set_active_block(True)
    self.widgets.lumabutton.set_active_block(False)
    self.widgets.shadowbutton.set_active_block(False)
    self.widgets.highlightbutton.set_active_block(False)
    self.widgets.diffbutton.set_active_block(False)
    self.widgets.minscale.set_sensitive(True)
    self.widgets.maxscale.set_sensitive(True)
    self.widgets.diffbutton.set_sensitive(self.refkey is not None and len(self.images) > 1)
    self.tabs.unblock_all_signals()
    self.tabs.set_current_page(0)
    self.window.show_all()

  def append_image(self, key, image):
    """Append a new tab for image 'image' with key 'key'.
       This can be done only once set_images has been called."""
    if self.images is None:
      raise RuntimeError("The method 'set_images' must be called before 'append_image'.")
    if key in self.images.keys():
      raise KeyError(f"The key '{key}' is already registered.")
    self.tabs.block_all_signals()
    self.images[key] = image.ref()
    self.images[key]._luma_ = self.images[key].luma()
    label = self.images[key].meta.get("tag", key)
    self.tabs.append_page(Gtk.Alignment(), Label(label)) # Append a zero size dummy child.
    self.tabs.unblock_all_signals()
    self.widgets.diffbutton.set_sensitive(self.refkey is not None)
    self.window.show_all()

  def update_image(self, key, image, create = False):
    """Update image with key 'key'.
       A new tab is appended if 'key' does not exist and 'create' is True. Otherwise, a KeyError exception is raised."""
    if key in self.images.keys():
      self.close_key_windows(key)
      self.images[key] = image.ref()
      self.images[key]._luma_ = self.images[key].luma()
      currentkey = self.get_current_key()
      redraw = (currentkey == key)
      if self.refkey == key:
        shadow = self.widgets.shadowbutton.get_active()
        highlight = self.widgets.highlightbutton.get_active()
        diff = self.widgets.diffbutton.get_active()
        redraw = redraw or shadow or highlight or diff
      if redraw: self.draw_image(currentkey)
    else:
      if create:
        self.append_image(key, image)
      else:
        raise KeyError(f"There is no image with key '{key}'.")

  def delete_image(self, key, force = False, failsafe = False):
    """Delete image with key 'key' if image.meta["deletable"] is False or 'force' is True.
       A KeyError exception is raised if image 'key' does not exist unless 'failsafe' is True."""
    try:
      image = self.images[key]
    except KeyError:
      if not failsafe: raise KeyError(f"There is no image with key '{key}'.")
      return
    deletable = image.meta.get("deletable", False)
    if not deletable and not force: return
    self.close_key_windows(key)
    if self.refkey == key:
      self.widgets.diffbutton.set_active_block(False)
      self.refkey = None
    self.tabs.block_all_signals()
    tab = list(self.images.keys()).index(key)
    del self.images[key]
    self.tabs.remove_page(tab)
    self.draw_image(self.get_current_key())
    self.tabs.unblock_all_signals()
    self.widgets.diffbutton.set_sensitive(self.refkey is not None and len(self.images) > 1)
    self.window.show_all()

  def get_nbr_images(self):
    """Return the number of image tabs."""
    return self.tabs.get_n_pages()

  def next_image(self, *args, **kwargs):
    """Show next image."""
    if self.images is None: return
    tab = (self.tabs.get_current_page()+1)%self.tabs.get_n_pages()
    self.tabs.set_current_page(tab)

  def previous_image(self, *args, **kwargs):
    """Show previous image."""
    if self.images is None: return
    tab = (self.tabs.get_current_page()-1)%self.tabs.get_n_pages()
    self.tabs.set_current_page(tab)

  # Show image description.

  def show_description(self):
    """Open image description popup."""
    if self.popup is not None: return
    key = self.get_current_key()
    description = self.images[key].meta.get("description", None)
    if description is None: return
    self.popup = Gtk.Window(Gtk.WindowType.POPUP, transient_for = self.window)
    self.popup.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
    self.popup.set_size_request(480, -1)
    label = Label(description, margin = 8)
    label.set_line_wrap(True)
    self.popup.add(label)
    self.popup.resize(1, 1)
    self.popup.show_all()

  def hide_description(self):
    """Close image description popup."""
    try:
      self.popup.destroy()
    except:
      pass
    self.popup = None

  # Show image statistics & light curve.

  def show_statistics(self):
    """Open image statistics window."""
    key = self.get_current_key()
    if key is None: return
    image = self.images[key]
    ax = self.canvas.figure.axes[0]
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    cropped = image.crop(np.ceil(xlim[0]), np.ceil(xlim[1]), np.ceil(ylim[1]), np.ceil(ylim[0]), inplace = False)
    self.statswindow.open(cropped, key, image.meta.get("tag", key))

  def show_lightcurve(self):
    """Open light curve window."""
    key = self.get_current_key()
    if key is None: return
    image = self.images[key]
    self.lightwindow.open(image, key, image.meta.get("tag", key))

  def close_key_windows(self, key = None):
    """Close statistics and light curve windows opened for key 'key' (any if None)."""
    self.statswindow.close(key = key)
    self.lightwindow.close(key = key)

  # Copy/paste callbacks.

  def set_copy_paste_callbacks(self, copy, paste):
    """Call 'copy(key, image)' (if not None) upon Ctrl+C, and 'paste(key, image)' (if not None) upon Ctrl+V,
       where 'image' is the image with key 'key'."""
    self.copy_callback = copy
    self.paste_callback = paste

  # Manage key & mouse button press/release events.

  def key_press(self, widget, event):
    """Callback for key press in the main window."""
    kbrd = decode_key(event)
    if kbrd.alt: return
    if kbrd.ctrl:
      key = self.get_current_key()
      if key is None: return
      if kbrd.uname == "C" and self.copy_callback is not None:
        self.copy_callback(key, self.images[key])
      elif kbrd.uname == "V" and self.paste_callback is not None:
        self.paste_callback(key, self.images[key])
      elif kbrd.uname == "X":
        self.delete_image(key)
      elif kbrd.uname == "TAB":
        if self.app.toolwindow.opened: self.app.toolwindow.window.present()
    else:
      if kbrd.uname == "PAGE_UP":
        self.previous_image()
      elif kbrd.uname == "PAGE_DOWN":
        self.next_image()
      elif kbrd.uname == "D":
        self.show_description()
      elif kbrd.uname == "S":
        self.show_statistics()

  def key_release(self, widget, event):
    """Callback for key release in the main window."""
    kbrd = decode_key(event)
    if kbrd.uname == "D":
      self.hide_description()

  def button_press(self, widget, event):
    """Callback for mouse button press in the main window."""
    if self.widgets.toolbar.mode != "": return # Don't mess with toolbar actions.
    if event.button == 3:
      self.contextmenu.popup_at_pointer(event)

  # Update luma RGB components.

  def get_rgb_luma(self):
    """Get luma RGB components."""
    return imageprocessing.get_rgb_luma()

  def rgb_luma_string(self, rgbluma = None):
    """Return luma RGB components 'rgbluma' as a string.
       If 'rgbluma' is None, get the current luma RGB components from self.get_rgb_luma()."""
    if rgbluma is None: rgbluma = self.get_rgb_luma()
    return f"Luma = {rgbluma[0]:.2f}R+{rgbluma[1]:.2f}G+{rgbluma[2]:.2f}B"

  def set_rgb_luma_callback(self, callback):
    """Call 'callback(rgbluma)' (if not None) upon update of the luma RGB components rgbluma."""
    self.rgb_luma_callback = callback

  def set_rgb_luma(self, rgbluma):
    """Set luma RGB components 'rgbluma'."""
    imageprocessing.set_rgb_luma(rgbluma)
    self.widgets.lumabutton.set_label(self.rgb_luma_string(rgbluma))
    for key in self.images.keys():
      self.images[key]._luma_ = self.images[key].luma()
    if self.widgets.lumabutton.get_active(): self.draw_image(self.get_current_key())
    if self.rgb_luma_callback is not None: self.rgb_luma_callback(rgbluma)

  def lock_rgb_luma(self):
    """Lock luma RGB components (disable Set button)."""
    self.widgets.rgblumabutton.set_sensitive(False)

  def unlock_rgb_luma(self):
    """Unlock luma RGB components (enable Set button)."""
    self.widgets.rgblumabutton.set_sensitive(True)

  # Guide lines.

  def set_guide_lines(self, plot_guide_lines, redraw = True):
    """Remove any existing guidelines and plot new ones with the method 'plot_guide_lines'.
       If not None, plot_guide_lines(ax) shall plot the guidelines in axes 'ax' and collect them in ax.guidelines.
       The main window canvas is redrawn if 'redraw' if True."""
    ax = self.canvas.figure.axes[0]
    try:
      for guideline in ax.guidelines: guideline.remove()
      del(ax.guidelines)
    except:
      pass
    self.plot_guide_lines = plot_guide_lines
    if self.plot_guide_lines is not None: self.plot_guide_lines(ax)
    if redraw:
      self.canvas.draw_idle()
      self.window.queue_draw()

  # Show activity.

  def set_busy(self):
    """Show the main window as busy."""
    self.widgets.spinner.start()
    #self.widgets.toolbar.set_message("Updating...")
    self.window.get_root_window().set_cursor(Gdk.Cursor(Gdk.CursorType.WATCH))

  def set_idle(self):
    """Show the main window as idle."""
    self.widgets.spinner.stop()
    #self.widgets.toolbar.set_message("")
    self.window.get_root_window().set_cursor(Gdk.Cursor(Gdk.CursorType.ARROW))



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Butterworth filter tool."""

from ..gtk.customwidgets import HBox, VBox, CheckButton, SpinButton, HScaleSpinButton
from ..toolmanager import BaseToolWindow
from skimage.filters import butterworth

class ButterworthFilterTool(BaseToolWindow):
  """Butterworth filter tool class."""

  _action_ = "Applying Butterworth filter..."

  _help_ = """Butterworth low-pass filter in the frequency domain:

    H(f) = 1/(1+(f/fc^(2n))

where n is the order of the filter and fc = (1-c)fs/2 is the cut-off frequency, with fs the sampling frequency and c\u2208[0, 1] the normalized cut-off frequency.
If the filter leaves visible artifacts on the edges, the image may be padded (with the edge values) prior to Fourier transform.
Apply the high-pass filter 1-H(f) if the high-pass checkbox is ticked."""

  _onthefly_ = False # This transformation can not be applied on the fly.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Butterworth filter"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.cutoffscale = HScaleSpinButton(.5, 0., .9999, .0001, digits = 4, length = 480)
    wbox.pack(self.widgets.cutoffscale.layout2("Normalized cut-off frequency:"))
    self.widgets.orderscale = HScaleSpinButton(2., .1, 10., .1, digits = 1, length = 480)
    wbox.pack(self.widgets.orderscale.layout2("Order:"))
    hbox = HBox()
    wbox.pack(hbox)
    self.widgets.padspin = SpinButton(0., 0., 32., 1., digits = 0)
    hbox.pack(self.widgets.padspin.hbox(prepend = "Padding:", append = "pixels"), expand = True, fill = True)
    self.widgets.highpassbutton = CheckButton(label = "High-pass")
    hbox.pack(self.widgets.highpassbutton)
    wbox.pack(self.tool_control_buttons())
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.cutoffscale.get_value(), self.widgets.orderscale.get_value(), int(round(self.widgets.padspin.get_value())), self.widgets.highpassbutton.get_active()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    cutoff, order, npad, highpass = params
    self.widgets.cutoffscale.set_value(cutoff)
    self.widgets.orderscale.set_value(order)
    self.widgets.padspin.set_value(npad)
    self.widgets.highpassbutton.set_active(highpass)

  def run(self, params):
    """Run tool for parameters 'params'."""
    cutoff, order, npad, highpass = params
    self.image.rgb = butterworth(self.reference.rgb, channel_axis = 0, cutoff_frequency_ratio = (1.-cutoff)/2., order = order, npad = npad, high_pass = highpass, squared_butterworth = True)
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    cutoff, order, npad, highpass = params
    band = "HighPass" if highpass else ""
    padding = f", padding = {npad} pixels" if npad > 0 else ""
    return f"{band}ButterworthFilter(cutoff = {cutoff:.4f}, order = {order:.1f}{padding})"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Unsharp mask tool."""

from ..gtk.customwidgets import HBox, VBox, RadioButtons, HScaleSpinButton
from ..toolmanager import BaseToolWindow
from ...imageprocessing import imageprocessing
from skimage.filters import unsharp_mask

class UnsharpMaskTool(BaseToolWindow):
  """Unsharp mask tool class."""

  _action_ = "Unsharp masking..."

  _onthefly_ = False # This transformation can not be applied on the fly.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Unsharp mask"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.channelbuttons = RadioButtons(("RGB", "RGB"), ("V", "HSV value"), ("L", "Luma"))
    wbox.pack(self.widgets.channelbuttons.hbox(prepend = "Channel(s):"))
    self.widgets.radiusscale = HScaleSpinButton(5., 0., 20., .01, digits = 2, length = 480)
    wbox.pack(self.widgets.radiusscale.layout2("Radius (pixels):"))
    self.widgets.amountscale = HScaleSpinButton(1., 0., 5., .01, digits = 2, length = 480)
    wbox.pack(self.widgets.amountscale.layout2("Amount:"))
    wbox.pack(self.tool_control_buttons())
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.channelbuttons.get_selected(), self.widgets.radiusscale.get_value(), \
           self.widgets.amountscale.get_value(), imageprocessing.get_rgb_luma()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    channels, radius, amount, rgbluma = params
    self.widgets.channelbuttons.set_selected(channels)
    self.widgets.radiusscale.set_value(radius)
    self.widgets.amountscale.set_value(amount)

  def run(self, params):
    """Run tool for parameters 'params'."""
    channels, radius, amount, rgbluma = params
    if amount <= 0. or radius <= 0.: return params, False
    if channels == "RGB":
      self.image.rgb = unsharp_mask(self.reference.rgb, channel_axis = 0, radius = radius, amount = amount)
    else:
      ref = self.reference.value() if channels == "V" else self.reference.luma()
      img = unsharp_mask(ref, radius = radius, amount = amount)
      self.image.copy_image_from(self.reference)
      self.image.scale_pixels(ref, img)
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    channels, radius, amount, rgbluma = params
    if channels == "L": channels = f"L({rgbluma[0]:.2f}, {rgbluma[1]:.2f}, {rgbluma[2]:.2f})"
    return f"UnsharpMask({channels}, radius = {radius:.2f} pixels, amount = {amount:.2f})"



# This program is 0free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Image processing tools."""

import re
import os
import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
from scipy.signal import convolve2d
from .defs import *
from . import utils
from . import colors
from .colors import rgbluma, get_rgb_luma, set_rgb_luma

from PIL import Image as PILImage
if IMAGEIO:
  import imageio.v3 as iio
else:
  import skimage.io as skio
import astropy.io.fits as pyfits

class Image:
  """Image class.
     The RGB components are stored as (3, height, width) arrays of floats in the range [0, 1].
     Note: No particular color space is assumed. Practically, most RGB images are encoded in
     the non-linear sRGB color space. It is the responsibility of the user to make the color
     space transformations appropriate to his needs."""

  # Object constructors, getters & setters.

  def __init__(self, image = None, meta = {}, channels = 0, copy = False):
    """Initialize object with RGB image 'image' (with channel axis 'channels') and meta-data 'meta'.
       The meta-data is a dictionary (or any other container) of user-defined data.
       The image is copied if 'copy' is True, or referenced (if possible) if False."""
    self.set_image(image, channels = channels, copy = copy)
    self.meta = meta

  @classmethod
  def newImage(cls, self, image = None, meta = {}, channels = 0, copy = False):
    """Return a new instance with RGB image 'image' (with channel axis 'channels') and meta-data 'meta'.
       The meta-data is a dictionary (or any other container) of user-defined data.
       The image is copied if 'copy' is True, or referenced (if possible) if False."""
    return cls(image = image, meta = meta, channels = channels, copy = copy)

  def set_image(self, image, channels = 0, copy = False):
    """Set RGB image 'image' (with channel axis 'channels') and return the object.
       The image is copied if 'copy' is True, or referenced (if possible) if False."""
    if image is None: # Short-circuit if image is None.
      self.rgb = None
      return self
    if not isinstance(image, np.ndarray):
      raise TypeError("Error, the image must be a numpy ndarray.")
    if image.ndim != 3:
      raise ValueError("Error, the image must be a 3D array.")
    if image.shape[channels] != 3:
      raise ValueError("Error, there must be exactly three (RGB) channels in the image.")
    if copy:
      self.rgb = IMGTYPE(np.copy(np.moveaxis(image, channels, 0)))
    else:
      self.rgb = IMGTYPE(np.moveaxis(image, channels, 0))
    return self

  def get_image(self):
    """Return a reference to the RGB image."""
    return self.rgb

  def get_image_copy(self):
    """Return a copy of the RGB image."""
    return self.rgb.copy()

  def set_meta(self, meta):
    """Set image meta-data 'meta' and return the object."""
    self.meta = meta
    return self

  def update_meta(self, newmeta):
    """Update image meta-data with 'newmeta' and return the object.
       Meaningful only if the image meta-data is a dictionary or any other container with an 'update' method."""
    self.meta.update(newmeta)
    return self

  def get_meta(self):
    """Return a reference to the image meta-data."""
    return self.meta

  # Object inquiries.

  def size(self):
    """Return the image width and height in pixels."""
    return self.rgb.shape[2], self.rgb.shape[1]

  def rgbf_view(self):
    """Return *a view* of the RGB components as a (height, width, 3) array of floats."""
    return np.moveaxis(self.rgb, 0, -1)

  def rgbf_copy(self):
    """Return *a copy* of the RGB components as a (height, width, 3) array of floats."""
    return np.moveaxis(self.rgb, 0, -1).copy()

  def rgb8(self):
    """Return the RGB components as a (height, width, 3) array of 8 bits integers in the range [0, 255]."""
    data = np.clip(self.rgb*255, 0, 255)
    return np.moveaxis(np.rint(data).astype("uint8"), 0, -1)

  def rgb16(self):
    """Return the RGB components as a (height, width, 3) array of 16 bits integers in the range [0, 65535]."""
    data = np.clip(self.rgb*65535, 0, 65535)
    return np.moveaxis(np.rint(data).astype("uint16"), 0, -1)

  def rgb32(self):
    """Return the RGB components as a (height, width, 3) array of 32 bits integers in the range [0, 4294967295]."""
    data = np.clip(self.rgb*4294967295, 0, 4294967295)
    return np.moveaxis(np.rint(data).astype("uint32"), 0, -1)

  def value(self):
    """Return the HSV value = max(RGB)."""
    return colors.hsv_value(self.rgb)

  def saturation(self):
    """Return the HSV saturation = 1-min(RGB)/max(RGB)."""
    return colors.hsv_saturation(self.rgb)

  def luma(self):
    """Return the luma."""
    return colors.luma(self.rgb)

  def luma16(self):
    """Return the luma as 16 bits integers in the range [0, 65535]."""
    data = np.clip(self.luma()*65535, 0, 65535)
    return np.rint(data).astype("uint16")

  def rgb_to_hsv(self):
    """Return the hue/saturation/value (HSV) components as a (height, width, 3) array of floats."""
    return colors.rgb_to_hsv(self.rgb)

  def set_hsv_image(self, hsv):
    """Set RGB image from the hue/saturation/value (HSV) data hsv(height, width, 3)."""
    self.rgb = colors.hsv_to_rgb(hsv)

  def srgb_to_lrgb(self):
    """Return the linear RGB components of a sRGB image."""
    return colors.srgb_to_lrgb(self.rgb)

  def lrgb_to_srgb(self):
    """Return the sRGB components of a linear RGB image."""
    return colors.lrgb_to_srgb(self.rgb)

  def srgb_luminance(self):
    """Return the luminance Y of a sRGB image."""
    return colors.srgb_luminance(self.rgb)

  def srgb_lightness(self):
    """Return the CIE lightness L* of a sRGB image."""
    return colors.srgb_lightness(self.rgb)

  def is_valid(self):
    """Return True if the object contains a valid RGB image, False otherwise."""
    return utils.is_valid_rgb_image(self.rgb)

  def is_out_of_range(self):
    """Return True if the RGB image is out-of-range (data < 0 or > 1 in any channel), False otherwise."""
    return np.any(self.rgb < -IMGTOL) or np.any(self.rgb > 1.+IMGTOL)

  def is_gray_scale(self):
    """Return True if the image is a gray scale (same RGB channels), False otherwise."""
    return np.all(abs(self.rgb[1]-self.rgb[0]) < IMGTOL) and np.all(abs(self.rgb[2]-self.rgb[0]) < IMGTOL)

  # Object copies.

  def ref(self, meta = "self"):
    """Return a new Image object with a reference to the RGB image and new meta-data 'meta' (copy of the original if meta = "self")."""
    if meta == "self": meta = deepcopy(self.meta)
    return self.newImage(self, self.rgb, meta, copy = False)

  def clone(self, meta = "self"):
    """Return a new Image object with a copy of the RGB image and new meta-data 'meta' (copy of the original if meta = "self")."""
    if meta == "self": meta = deepcopy(self.meta)
    return self.newImage(self, self.rgb, meta, copy = True)

  def copy_image_from(self, source):
    """Copy the RGB image from 'source'."""
    self.set_image(source.rgb, copy = True)

  def copy_meta_from(self, source):
    """Copy the meta-data from 'source'."""
    self.meta = deepcopy(source.meta)

  # Image load/save.

  def load(self, filename, meta = {}):
    """Load file 'filename' and set meta-data 'meta' (leave unchanged if meta = "self", or pick the file meta-data if meta = "file").
       Return the file meta-data (including exif if available). The image color space is not transformed and assumed to be sRGB."""
    print(f"Loading file {filename}...")
    try:
      header = PILImage.open(filename)
      fmt = header.format
      print(f"Format = {fmt}.")
    except:
      header = None
      fmt = None
      print("Failed to identify image file format; Attempting to load anyway...")
    if fmt == "PNG": # Load with the FreeImage plugin to enable 16 bits color depth.
      image = iio.imread(filename, plugin = "PNG-FI") if IMAGEIO else skio.imread(filename)
    elif fmt == "TIFF":
      image = iio.imread(filename, plugin = "TIFF") if IMAGEIO else skio.imread(filename, plugin = "tifffile")
    elif fmt == "FITS":
      hdus = pyfits.open(filename)
      image = hdus[0].data
      if image.ndim == 3:                 # Pyfits returns (channels, height, width)
        image = np.moveaxis(image, 0, -1) # instead of (height, width, channels),
      image = np.flip(image, axis = 0)    # and an upside down image.
    else:
      image = iio.imread(filename) if IMAGEIO else skio.imread(filename)
    if image.ndim == 2:
      nc = 1
    elif image.ndim == 3:
      nc = image.shape[2]
    else:
      raise ValueError(f"Error, invalid image dimensions = {image.shape}.")
    print(f"Number of channels = {nc}.")
    if nc not in [1, 3, 4]:
      raise ValueError(f"Error, images with {nc} channels are not supported.")
    print(f"Image size = {image.shape[1]}x{image.shape[0]} pixels.")
    dtype = str(image.dtype)
    print(f"Data type = {dtype}.")
    if dtype == "uint8":
      bpc = 8
      image = IMGTYPE(image/255)
    elif dtype == "uint16":
      bpc = 16
      image = IMGTYPE(image/65535)
    elif dtype == "uint32":
      bpc = 32
      image = IMGTYPE(image/4294967295)
    elif dtype in ["float32", ">f4", "<f4"]: # Assumed normalized in [0, 1] !
      bpc = 32
      image = IMGTYPE(image)
    elif dtype in ["float64", ">f8", "<f8"]: # Assumed normalized in [0, 1] !
      bpc = 64
      image = IMGTYPE(image)
    else:
      raise TypeError(f"Error, image data type {dtype} is not supported.")
    print(f"Bit depth per channel = {bpc}.")
    print(f"Bit depth per pixel = {nc*bpc}.")
    if nc == 1: # Assume single channel images are monochrome.
      image = np.repeat(image[:, :, np.newaxis], 3, axis = 2)
    image = np.moveaxis(image, -1, 0) # Move last (channel) axis to leading position.
    for ic in range(nc):
      print(f"Channel #{ic}: minimum = {image[ic].min():.5f}, maximum = {image[ic].max():.5f}.")
    if nc == 4: # Assume fourth channel is transparency.
      image = image[0:3]*image[3]
    self.rgb = np.ascontiguousarray(image)
    try:
      exif = header.getexif()
      print("Succesfully read EXIF data...")
    except:
      exif = None
    filemeta = {"exif": exif, "colordepth": bpc}
    #print(f"File meta-data = {filemeta}.")
    if meta == "file":
      self.meta = deepcopy(filemeta)
    elif meta != "self":
      self.meta = meta
    return filemeta

  def save(self, filename, depth = 8, single_channel_gray_scale = True):
    """Save image in file 'filename' with color depth 'depth' (bits/channel).
       The file format is chosen according to the 'filename' extension:
        - .png : PNG file with depth = 8 or 16 bits/channel.
        - .tif, .tiff : TIFF file with depth = 8, 16 (integers), or 32 (floats) bits/channel.
        - .fit/.fits/.fts : FITS file with 32 bits (floats)/channel (irrespective of depth).
       The image is saved as a single channel gray scale if all RGB channels are the same and 'single_channel_gray_scale' is True.
       The image color space is assumed to be sRGB."""
    is_gray_scale = single_channel_gray_scale and self.is_gray_scale()
    if is_gray_scale:
      print(f"Saving gray scale image as file {filename}...")
    else:
      print(f"Saving RGB image as file {filename}...")
    root, ext = os.path.splitext(filename)
    if ext in [".png", ".tif", ".tiff"]:
      if depth == 8:
        image = self.rgb8()
      elif depth == 16:
        image = self.rgb16()
      elif depth == 32:
        image = self.rgb32()
      else:
        raise ValueError("Error, color depth must be 8 or 16, or 32 bits.")
      print(f"Color depth = {depth} bits per channel (integers).")
      if is_gray_scale: image = image[:, : , 0]
      if ext == ".png":
        if IMAGEIO:
          if depth > 16: raise ValueError("Error, color depth of png files must be 8 or 16 bits per channel.")
          iio.imwrite(filename, image, plugin = "PNG-FI")
        else:
          if depth > 8: raise ValueError("Error, color depth of png files must be 8 bits per channel.")
          skio.imsave(filename, image, check_contrast = False)
      elif ext == ".tif" or ext == ".tiff":
        if IMAGEIO:
          iio.imwrite(filename, image, plugin = "TIFF", metadata = {"compress": 5})
        else:
          skio.imsave(filename, image, plugin = "tifffile", check_contrast = False, compression = "zlib")
    elif ext in [".fit", ".fits", ".fts"]:
      print(f"Color depth = {np.finfo(IMGTYPE).bits} bits per channel (floats).")
      if is_gray_scale:
        image = np.flip(self.rgb[0], axis = 0)
      else:
        image = np.flip(self.rgb, axis = 1)
      hdu = pyfits.PrimaryHDU(image)
      hdu.writeto(filename, overwrite = True)
    else:
      raise ValueError("Error, file extension must be .png or .tif/.tiff., or .fit/.fits/.fts.")

  # Image draw.

  def draw(self, ax):
    """Draw the image in matplotlib axes 'ax'.
       The image color space is assumed to be sRGB."""
    ax.imshow(self.rgbf())

  # Image statistics & histograms.

  def statistics(self, channels = "RGBVL"):
    """Compute image statistics for channels 'channels', a combination of the keys "R" (for red), "G" (for green), "B" (for blue),
       "V" (for HSV value), "L" (for luma) and "S" (for HSV saturation). Return stats[key] for key in channels, with:
         - stats[key].name = channel name ("Red", "Green", "Blue", "Value", "Luma" or "Saturation", provided for convenience).
         - stats[key].width = image width (provided for convenience).
         - stats[key].height = image height (provided for convenience).
         - stats[key].npixels = number of image pixels = image width*image height (provided for convenience).
         - stats[key].minimum = minimum value in channel key.
         - stats[key].maximum = maximum value in channel key.
         - stats[key].percentiles = (pr25, pr50, pr75) = the 25th, 50th and 75th percentiles in channel key (excluding pixels <= 0 and >= 1).
         - stats[key].median = pr50 = median value in channel key (excluding pixels <= 0 and >= 1).
         - stats[key].zerocount = number of pixels <= 0 in channel key.
         - stats[key].oorcount = number of pixels  > 1 (out-of-range) in channel key."""
    class Container: pass # An empty container class.
    width, height = self.size()
    npixels = width*height
    stats = {}
    for key in channels:
      if key == "R":
        name = "Red"
        channel = self.rgb[0]
      elif key == "G":
        name = "Green"
        channel = self.rgb[1]
      elif key == "B":
        name = "Blue"
        channel = self.rgb[2]
      elif key == "V":
        name = "Value"
        channel = self.value()
      elif key == "L":
        name = "Luma"
        channel = self.luma()
      elif key == "S":
        name = "Saturation"
        channel = self.saturation()
      else:
        raise ValueError(f"Error, invalid channel '{key}'.")
      stats[key] = Container()
      stats[key].name = name
      stats[key].width = width
      stats[key].height = height
      stats[key].npixels = npixels
      stats[key].minimum = channel.min()
      stats[key].maximum = channel.max()
      mask = (channel >= IMGTOL) & (channel <= 1.-IMGTOL)
      if np.any(mask):
        stats[key].percentiles = np.percentile(channel[mask], [25., 50., 75.])
        stats[key].median = stats[key].percentiles[1]
      else:
        stats[key].percentiles = None
        stats[key].median = None
      stats[key].zerocount = np.sum(channel < IMGTOL)
      stats[key].outcount = np.sum(channel > 1.+IMGTOL)
    return stats

  def histograms(self, channels = "RGBVL", nbins = 256):
    """Return image histograms for channels 'channels', a combination of the keys "R" (for red), "G" (for green), "B" (for blue),
       "V" (for HSV value), "L" (for luma) and "S" (for HSV saturation). Return a tuple (edges, counts), where edges(nbins) are
       the bin edges and counts(len(channels), nbins) are the bin counts for all channels. 'nbins' is the number of bins in the
       range [0, 1]."""
    minimum = min(0., self.rgb.min())
    maximum = max(1., self.rgb.max())
    nbins = int(round(nbins*(maximum-minimum)))
    counts = np.empty((len(channels), nbins))
    ic = 0
    for key in channels:
      if key == "R":
        channel = self.rgb[0]
      elif key == "G":
        channel = self.rgb[1]
      elif key == "B":
        channel = self.rgb[2]
      elif key == "V":
        channel = self.value()
      elif key == "L":
        channel = self.luma()
      elif key == "S":
        channel = self.saturation()
      else:
        raise ValueError(f"Error, invalid channel '{key}'.")
      counts[ic], edges = np.histogram(channel, bins = nbins, range = (minimum, maximum), density = False)
      ic += 1
    return edges, counts

  ##########################
  # Image transformations. #
  ##########################

  # Image normalization.

  def clip(self):
    """Clip the image in the [0, 1] range."""
    self.rgb = np.clip(self.rgb, 0., 1.)

  def scale_pixels(self, source, target):
    """Scale all pixels of the image by the ratio target/source.
       Wherever abs(source) < IMGTOL, set all channels to target."""
    self.rgb = utils.scale_pixels(self.rgb, source, target)

  def protect_highlights(self, luma = None):
    """Normalize out-of-range pixels with HSV value > 1 by adjusting the saturation at constant luma.
       'luma' is the luma of the image, if available (if None, the luma is recomputed on the fly).
       Warning: This method aims at protecting the highlights from overflowing when stretching the luma.
       It assumes that the luma remains <= 1 even though some pixels have HSV value > 1."""
    if luma is None: luma = self.luma() # Original luma.
    self.rgb /= np.maximum(self.rgb.max(axis = 0), 1.) # Rescale maximum HSV value to 1.
    newluma = self.luma() # Updated luma.
    # Scale the saturation.
    # Note: The following implementation is failsafe when newluma -> 1 (in which case luma is also 1 in principle),
    # at the cost of a small error.
    fs = ((1.-luma)+IMGTOL)/((1.-newluma)+IMGTOL)
    self.rgb = 1.-fs*(1.-self.rgb)

  # Histogram transformations.

  def clip_shadows_highlights(self, shadow = None, highlight = None, channels = "V", inplace = True, meta = "self"):
    """Clip channels 'channels' below shadow level 'shadow' and above highlight level 'highlight', and
       remap [shadow, highlight] to [0, 1]. 'channels' can be "V" (value), "L" (luma) or any combination
       of "R" (red), "G" (green), and "B" (blue). shadow = min(channel) for each channel if 'shadow' is none,
       and highlight = max(channel) for each channel if 'highlight' is None. Also set new meta-data 'meta'
       (same as the original if meta = "self"). Update the object if 'inplace' is True or return a new instance
       if 'inplace' is False."""
    if channels in ["V", "L"]:
      channel = self.value() if channels == "V" else self.luma()
      if shadow is None: shadow = max(channel.min(), 0.)
      if highlight is None: highlight = channel.max()
      clipped = np.clip(channel, shadow, highlight)
      interpd = np.interp(clipped, (shadow, highlight), (0., 1.))
      image = utils.scale_pixels(self.rgb, channel, interpd)
      if inplace: self.rgb = image
    else:
      image = self.rgb if inplace else self.rgb.copy()
      for ic, key in ((0, "R"), (1, "G"), (2, "B")):
        if key in channels:
          shadow_ = max(image[ic].min(), 0.) if shadow is None else shadow
          highlight_ = image[ic].max() if highlight is None else highlight
          clipped = np.clip(image[ic], shadow_, highlight_)
          image[ic] = np.interp(clipped, (shadow_, highlight_), (0., 1.))
    if inplace:
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  def set_dynamic_range(self, fr = None, to = (0., 1.), channels = "L", inplace = True, meta = "self"):
    """Remap 'channels' from range 'fr' (a tuple) to range 'to' (a tuple, default (0, 1)).
       'channels' can be "V" (value), "L" (luma) or any combination of "R" (red) "G" (green), and "B" (blue).
       fr = (min(channel), max(channel)) for each channel if 'fr' is None. Also set new meta-data 'meta'
       (same as the original if meta = "self"). Update the object if 'inplace' is True or return a new
       instance if 'inplace' is False."""
    if fr is not None:
      if fr[1] <= fr[0]: raise ValueError("Error, fr[1] must be > fr[0].")
    if to[1] <= to[0]: raise ValueError("Error, to[1] must be > to[0].")
    if channels in ["V", "L"]:
      channel = self.value() if channels == "V" else self.luma()
      fr_ = (channel.min(), channel.max()) if fr is None else fr
      interpd = np.maximum(np.interp(channel, fr_, to), 0.)
      image = utils.scale_pixels(self.rgb, channel, interpd)
      if inplace: self.rgb = image
    else:
      image = self.rgb if inplace else self.rgb.copy()
      for ic, key in ((0, "R"), (1, "G"), (2, "B")):
        if key in channels:
          fr_ = (image[ic].min(), image[ic].max()) if fr is None else fr
          image[ic] = np.maximum(np.interp(image[ic], fr_, to), 0.)
    if inplace:
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  def gamma_correction(self, gamma, channels = "L", inplace = True, meta = "self"):
    """Apply gamma correction with exponent 'gamma' to channels 'channels'.
       'channels' can be "V" (value), "L" (luma) or any combination of "R" (red) "G" (green), and "B" (blue).
       Also set new meta-data 'meta' (same as the original if meta = "self"). Update the object if 'inplace'
       is True or return a new instance if 'inplace' is False."""
    if gamma <= 0.: raise ValueError("Error, gamma must be >= 0.")
    if channels in ["V", "L"]:
      channel = self.value() if channels == "V" else self.luma()
      clipped = np.clip(channel, 0., 1.)
      corrected = clipped**gamma
      image = utils.scale_pixels(self.rgb, channel, corrected)
      if inplace: self.rgb = image
    else:
      image = self.rgb if inplace else self.rgb.copy()
      for ic, key in ((0, "R"), (1, "G"), (2, "B")):
        if key in channels:
          clipped = np.clip(image[ic], 0., 1.)
          image[ic] = clipped**gamma
    if inplace:
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  def midtone_correction(self, midtone = .5, channels = "L", inplace = True, meta = "self"):
    """Apply midtone correction with midtone 'midtone' to channels 'channels'.
       'channels' can be "V" (value), "L" (luma) or any combination of "R" (red) "G" (green), and "B" (blue).
       Also set new meta-data 'meta' (same as the original if meta = "self"). Update the object if 'inplace'
       is True or return a new instance if 'inplace' is False."""
    if midtone <= 0.: raise ValueError("Error, midtone must be >= 0.")
    if channels in ["V", "L"]:
      channel = self.value() if channels == "V" else self.luma()
      clipped = np.clip(channel, 0., 1.)
      stretched = (midtone-1.)*clipped/((2.*midtone-1.)*clipped-midtone)
      image = utils.scale_pixels(self.rgb, channel, stretched)
      if inplace: self.rgb = image
    else:
      image = self.rgb if inplace else self.rgb.copy()
      for ic, key in ((0, "R"), (1, "G"), (2, "B")):
        if key in channels:
          clipped = np.clip(image[ic], 0., 1.)
          image[ic] = (midtone-1.)*clipped/((2.*midtone-1.)*clipped-midtone)
    if inplace:
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  def generalized_stretch(self, stretch_function, params, channels = "L", inplace = True, meta = "self"):
    """Stretch histogram of channels 'channels' with an arbitrary stretch function 'stretch_function' parametrized
       by 'params'. The function stretch_function(input, params) shall return the output levels for an array of input
       levels 'input'. 'channels' can be "V" (value), "L" (luma) or any combination of "R" (red) "G" (green),
       and "B" (blue). Also set new meta-data 'meta' (same as the original if meta = "self"). Update the object
       if 'inplace' is True or return a new instance if 'inplace' is False."""
    if channels in ["V", "L"]:
      channel = self.value() if channels == "V" else self.luma()
      stretched = IMGTYPE(stretch_function(channel, params))
      image = utils.scale_pixels(self.rgb, channel, stretched)
      if inplace: self.rgb = image
    else:
      image = self.rgb if inplace else self.rgb.copy()
      for ic, key in ((0, "R"), (1, "G"), (2, "B")):
        if key in channels:
          image[ic] = IMGTYPE(stretch_function(image[ic], params))
    if inplace:
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  def generalized_stretch_lookup(self, stretch_function, params, channels = "L", inplace = True, meta = "self", nlut = 131072):
    """Stretch histogram of channels 'channels' with an arbitrary stretch function 'stretch_function' parametrized
       by 'params'. The function stretch_function(input, params) shall return the output levels for an array of input
       levels 'input'. 'channels' can be "V" (value), "L" (luma) or any combination of "R" (red) "G" (green),
       and "B" (blue). Also set new meta-data 'meta' (same as the original if meta = "self"). Update the object
       if 'inplace' is True or return a new instance if 'inplace' is False.
       This method uses a look-up table with linear interpolation between 'nlut' elements to apply the stretch function
       to the channel(s); It shall be much faster than generalized_stretch(...) when the stretch function is expensive.
       The original image is clipped in the [0, 1] range before stretching."""
    xlut = np.linspace(0., 1., nlut, dtype = IMGTYPE) # Build the look-up table.
    ylut = IMGTYPE(stretch_function(xlut, params))
    slut = (ylut[1:]-ylut[:-1])/(xlut[1:]-xlut[:-1]) # Slopes.
    if channels in ["V", "L"]:
      channel = self.value() if channels == "V" else self.luma()
      clipped = np.clip(channel, 0., 1.)
      stretched = utils.lookup(clipped, xlut, ylut, slut, nlut)
      image = utils.scale_pixels(self.rgb, channel, stretched)
      if inplace: self.rgb = image
    else:
      image = self.rgb if inplace else self.rgb.copy()
      for ic, key in ((0, "R"), (1, "G"), (2, "B")):
        if key in channels:
          clipped = np.clip(image[ic], 0., 1.)
          image[ic] = utils.lookup(clipped, xlut, ylut, slut, nlut)
    if inplace:
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  # Color transformations.

  def color_balance(self, red = 1., green = 1., blue = 1., inplace = True, meta = "self"):
    """Multiply the red channel by 'red', the green channel by 'green', and the blue channel by 'blue'.
       Also set new meta-data 'meta' (same as the original if meta = "self"). Update the object if 'inplace'
       is True or return a new instance if 'inplace' is False."""
    if red < 0.: raise ValueError("Error, red must be >= 0.")
    if green < 0.: raise ValueError("Error, green must be >= 0.")
    if blue < 0.: raise ValueError("Error, blue must be >= 0.")
    if inplace:
      if meta != "self": self.meta = meta
      image = self.rgb
    else:
      if meta == "self": meta = deepcopy(self.meta)
      image = self.rgb.copy()
    if red   != 1.: image[0] *= red
    if green != 1.: image[1] *= green
    if blue  != 1.: image[2] *= blue
    return None if inplace else self.newImage(self, image, meta)

  def negative(self, inplace = True, meta = "self"):
    """Make a negative of the image and set new meta-data 'meta' (same as the original if meta = "self").
       Update the object if 'inplace' is True or return a new instance if False."""
    image = np.clip(1.-self.rgb, 0., 1.)
    if inplace:
      self.rgb = image
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  def gray_scale(self, channel = "L", inplace = True, meta = "self"):
    """Convert into a gray scale using channel = "V" (HSV value), "L" (luma) or "Y" (luminance),
       and set new meta-data 'meta' (same as the original if meta = "self").
       Update the object if 'inplace' is True or return a new instance if False."""
    if inplace:
      if meta != "self": self.meta = meta
      image = self.rgb
    else:
      if meta == "self": meta = deepcopy(self.meta)
      image = np.empty_like(self.rgb)
    if channel == "V":
      image[:] = self.value()
    elif channel == "L":
      image[:] = self.luma()
    elif channel == "Y":
      image[:] = colors.lrgb_to_srgb(self.srgb_luminance())
    else:
      raise ValueError(f"Error, invalid channel '{channel}'.")
    return None if inplace else self.newImage(self, image, meta)

  # Image enhancement.

  def sharpen(self, inplace = True, meta = "self"):
    """Apply a sharpening convolution filter and set new meta-data 'meta' (same
       as the original if meta = "self"). Update the object if 'inplace' is True or
       return a new instance if 'inplace' is False."""
    if inplace:
      if meta != "self": self.meta = meta
      image = self.rgb
    else:
      if meta == "self": meta = deepcopy(self.meta)
      image = np.empty_like(self.rgb)
    kernel = np.array([[-1., -1., -1.], [-1., 9., -1.], [-1., -1., -1.]], dtype = IMGTYPE)
    for ic in range(3):
      image[ic] = convolve2d(self.rgb[ic], kernel, mode = "same", boundary = "fill", fillvalue = 0.)
    return None if inplace else self.newImage(self, image, meta)

  def remove_hot_pixels(self, ratio = 2., channels = "L", inplace = True, meta = "self"):
    """Remove hot pixels in channels 'channels'. 'channels' can be "V" (value), "L" (luma) or any
       combination of "R" (red), "G" (green), and "B" (blue). All pixels in a channel whose level is
       greater than 'ratio' times the average of their 8 nearest neighbors are replaced by this average.
       Also set new meta-data 'meta' (same as the original if meta = "self"). Update the object if 'inplace'
       is True or return a new instance if 'inplace' is False."""
    if ratio <= 0.: raise ValueError("Error, ratio must be > 0.")
    if inplace:
      if meta != "self": self.meta = meta
      image = self.rgb
    else:
      if meta == "self": meta = deepcopy(self.meta)
      image = np.empty_like(self.rgb)
    kernel = np.array([[1., 1., 1.], [1., 0., 1.], [1., 1., 1.]], dtype = IMGTYPE)
    if channels in ["V", "L"]:
      channel = self.value() if channels == "V" else self.luma()
      nnn = convolve2d(np.ones_like(channel), kernel, mode = "same", boundary = "fill", fillvalue = 0.)
      avg = convolve2d(channel, kernel, mode = "same", boundary = "fill", fillvalue = 0.)/nnn
      mask = (channel > ratio*avg)
      for channel in range(3):
        avg = convolve2d(self.rgb[channel], kernel, mode = "same", boundary = "fill", fillvalue = 0.)/nnn
        image[channel] = np.where(mask, avg, self.rgb[channel])
    else:
      nnn = convolve2d(np.ones_like(self.rgb[0]), kernel, mode = "same", boundary = "fill", fillvalue = 0.)
      for ic, key in ((0, "R"), (1, "G"), (2, "B")):
        if key in channels:
          avg = convolve2d(self.rgb[ic], kernel, mode = "same", boundary = "fill", fillvalue = 0.)/nnn
          image[ic] = np.where(self.rgb[ic] > ratio*avg, avg, self.rgb[ic])
    return None if inplace else self.newImage(self, image, meta)

  # Image resizing & crop.

  def resize(self, width, height, resample = LANCZOS, inplace = True, meta = "self"):
    """Resize image to width 'width' and height 'height' using resampling method 'resample'
       (either NEAREST, BILINEAR, BICUBIC, LANCZOS, BOX or HAMMING). Also set new meta-data
       'meta' (same as the original if meta = "self"). Update the object if 'inplace' is True
       or return a new instance if 'inplace' is False."""
    if width < 1 or width > 32768: raise ValueError("Error, width must be >= 1 and <= 32768 pixels.")
    if height < 1 or height > 32768: raise ValueError("Error, height must be >= 1 and <= 32768 pixels.")
    if width*height > 2**26: raise ValueError("Error, can not resize to > 64 Mpixels.")
    if not resample in [NEAREST, BILINEAR, BICUBIC, LANCZOS, BOX, HAMMING]: raise ValueError("Error, invalid resampling method.")
    image = np.empty((3, height, width), dtype = IMGTYPE)
    for ic in range(3): # Resize each channel using PIL.
      PILchannel = PILImage.fromarray(np.float32(self.rgb[ic]), "F").resize((width, height), resample) # Convert to np.float32 while resizing.
      image[ic] = np.asarray(PILchannel, dtype = IMGTYPE)
    if inplace:
      self.rgb = image
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  def rescale(self, scale, resample = LANCZOS, inplace = True, meta = "self"):
    """Rescale image by a factor 'scale' using resampling method 'resample' (either NEAREST,
       BILINEAR, BICUBIC, LANCZOS, BOX or HAMMING). Also set new meta-data 'meta' (same as
       the original if meta = "self"). Update the object if 'inplace' is True or return a new
       instance if 'inplace' is False."""
    if scale <= 0. or scale > 16.: raise ValueError("Error, scale must be > 0 and <= 16.")
    width, height = self.size()
    newwidth, newheight = int(round(scale*width)), int(round(scale*height))
    return self.resize(newwidth, newheight, resample, inplace, meta)

  def crop(self, xmin, xmax, ymin, ymax, inplace = True, meta = "self"):
    """Crop image from x = xmin to x = xmax and from y = ymin to y = ymax.
       Also set new meta-data 'meta' (same as the original if meta = "self").
       Update the object if 'inplace' is True or return a new instance if 'inplace' is False."""
    if xmax <= xmin: raise ValueError("Error, xmax <= xmin.")
    if ymax <= ymin: raise ValueError("Error, ymax <= ymin.")
    width, height = self.size()
    xmin = max(int(np.floor(xmin))  , 0)
    xmax = min(int(np.ceil (xmax))+1, width)
    ymin = max(int(np.floor(ymin))  , 0)
    ymax = min(int(np.ceil (ymax))+1, height)
    if inplace:
      self.rgb = self.rgb[:, ymin:ymax, xmin:xmax]
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, self.rgb[:, ymin:ymax, xmin:xmax], meta)

# Special images and shortcuts.

def black_image(width, height, meta = {}):
  """Return a black image with width 'width', height 'height', and meta-data 'meta'."""
  return Image(np.zeros((3, height, width), dtype = IMGTYPE), meta)

def white_image(width, height, meta = {}):
  """Return a white image with width 'width', height 'height', and meta-data 'meta'."""
  return Image(np.ones((3, height, width), dtype = IMGTYPE), meta)

def load_image(filename, meta = "self"):
  """Return the image in file 'filename' and set meta-data 'meta'."""
  image = Image()
  image.load(filename, meta)
  return image



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Image processing tools for Unistellar frames."""

import numpy as np
from copy import deepcopy
import matplotlib.pyplot as plt
from .imageprocessing import Image

Unistellar_warned_about_frames = False

class UnistellarImage(Image):
  """Image class for the Unistellar frames. Handles the Unistellar frame."""

  # type: Telescope & image type.
  # width: Framed image width.
  # height: Framed image height.
  # radius: Frame radius.
  # margin: Frame margin (between the frame & image).
  # cropradius: Crop radius (crop image at that radius to remove the frame).
  # threshold: Minimal HSV value of the frame.

  __FRAMES__ = [{"type": "eQuinox 1", "width": 2240, "height": 2240, "radius": 1050, "margin": 64, "cropradius": 996, "threshold": 24/255},
                {"type": "eQuinox 1 (Planets)", "width": 1120, "height": 1120, "radius": 525, "margin": 32, "cropradius": 498, "threshold": 24/255}]

  def check_frame(self):
    """Check if the image has an Unistellar frame.
       Return the dictionary of frame properties if so, None otherwise."""
    global Unistellar_warned_about_frames
    if not Unistellar_warned_about_frames:
      print("#########################################################################")
      print("# WARNING: Frame detection is presently based on the size of the image. #")
      print("#          This will (hopefully) be made more robust later.             #")
      print("#########################################################################")
      Unistellar_warned_about_frames = True
    for framed in self.__FRAMES__:
      if self.rgb.shape == (3, framed["height"], framed["width"]):
        return framed
    return None

  def draw_crop_boundary(self, ax = None, color = "yellow", linestyle = "--", linewidth = 1.):
    """Draw the Unistellar frame crop boundary in axes 'ax' (gca() if None) with linestyle 'linestyle', linewidth 'linewidth' and color 'color'."""
    framed = self.check_frame()
    if not framed: return
    width = framed["width"]
    height = framed["height"]
    radius = framed["cropradius"]
    if ax is None: ax = plt.gca()
    ax.add_patch(plt.Circle((width/2, height/2), radius, linestyle = linestyle, linewidth = linewidth, color = color, fill = False))

  def get_frame(self):
    """Return the Unistellar frame as an image."""
    framed = self.check_frame()
    if not framed: return None
    width = framed["width"]
    height = framed["height"]
    radius = framed["cropradius"]
    threshold = framed["threshold"]
    x = np.arange(0, width)-(width-1)/2
    y = np.arange(0, height)-(height-1)/2
    X, Y = np.meshgrid(x, y, sparse = True)
    outer = (X**2+Y**2 > radius**2)
    mask = outer & (self.value() >= threshold)
    frame = np.zeros_like(self.rgb)
    frame[:, mask] = self.rgb[:, mask]
    return Image(frame, {"description": "Unistellar Frame"})

  def remove_frame(self, frame, inplace = True, meta = "self"):
    """Remove the Unistellar frame from the image and set new meta-data 'meta' (same as the original if meta = "self").
       Update the object if 'inplace' is True or return a new instance if 'inplace' is False."""
    image = np.where(frame.value() > 0., 0., self.rgb)
    if inplace:
      self.rgb = image
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)

  def add_frame(self, frame, inplace = True, meta = "self"):
    """Add the Unistellar frame 'frame' to the image and set new meta-data 'meta' (same as the original if meta = "self").
       Update the object if 'inplace' is True or return a new instance if 'inplace' is False."""
    image = np.where(frame.value() > 0., frame.rgb, self.rgb)
    if inplace:
      self.rgb = image
      if meta != "self": self.meta = meta
      return None
    else:
      if meta == "self": meta = deepcopy(self.meta)
      return self.newImage(self, image, meta)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Wavelets filter tool."""

from ..gtk.customwidgets import Align, HBox, VBox, CheckButton, RadioButtons, SpinButton, ComboBoxText, Entry
from ..toolmanager import BaseToolWindow
from skimage.restoration import estimate_sigma, denoise_wavelet, cycle_spin

class WaveletsFilterTool(BaseToolWindow):
  """Wavelets filter tool class."""

  _action_ = "Applying wavelets filter..."

  _onthefly_ = False # This transformation can not be applied on the fly.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Wavelets filter"): return False
    sigma = estimate_sigma(self.reference.rgb, channel_axis = 0, average_sigmas = False)
    wbox = VBox()
    self.window.add(wbox)
    hbox = HBox()
    wbox.pack(hbox)
    hbox.pack("Estimated noise level in each channel:")
    self.widgets.bindbutton = CheckButton(label = "Bind RGB channels", halign = Align.END)
    self.widgets.bindbutton.connect("toggled", lambda button: self.update(0))
    hbox.pack(self.widgets.bindbutton, expand = True, fill = True)
    hbox = HBox()
    wbox.pack(hbox)
    self.widgets.entries = []
    for channel, label in ((0, "Red:"), (1, 8*" "+"Green:"), (2, 8*" "+"Blue:")):
      entry = Entry(text = f"{sigma[channel]:.5e}", width = 12)
      entry.channel = channel
      entry.connect("changed", lambda entry: self.update(entry.channel))
      self.widgets.entries.append(entry)
      hbox.pack(label)
      hbox.pack(entry)
    self.widgets.ycbcrbutton = CheckButton(label = "Apply transformation in the YCbCr color space")
    self.widgets.ycbcrbutton.set_active(True)
    wbox.pack(self.widgets.ycbcrbutton)
    wavelets = [] # Set-up the list of wavelets.
    for i in range(8):
      wavelets.append((f"db{i+1}", f"Daubechies {i+1}"))
    for i in range(7):
      wavelets.append((f"sym{i+2}", f"Symlets {i+2}"))
    for i in range(8):
      wavelets.append((f"coif{i+1}", f"Coiflets {i+1}"))
    self.widgets.waveletcombo = ComboBoxText(*wavelets)
    self.widgets.waveletcombo.set_selected("coif4")
    wbox.pack(self.widgets.waveletcombo.hbox(prepend = "Wavelets:"))
    hbox = HBox()
    wbox.pack(hbox)
    self.widgets.methodbuttons = RadioButtons(("BayesShrink", "Bayes Shrink"), ("VisuShrink", "Visu Shrink"))
    hbox.pack(self.widgets.methodbuttons.hbox(prepend = "Method:"))
    self.widgets.modebuttons = RadioButtons(("Soft", "Soft"), ("Hard", "Hard"))
    hbox.pack(self.widgets.modebuttons.hbox(prepend = "        Mode:"))
    self.widgets.shiftsbutton = SpinButton(0., 0., 8., 1., page = 1., digits = 0)
    wbox.pack(self.widgets.shiftsbutton.hbox(prepend = "Maximum shift for cycle spinning:"))
    wbox.pack(self.tool_control_buttons())
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    try:
      sigma = tuple(float(self.widgets.entries[channel].get_text()) for channel in range(3))
    except:
      return None
    return sigma, self.widgets.ycbcrbutton.get_active(), self.widgets.waveletcombo.get_selected(), self.widgets.methodbuttons.get_selected(), \
           self.widgets.modebuttons.get_selected(), int(self.widgets.shiftsbutton.get_value())

  def set_params(self, params):
    """Set tool parameters 'params'."""
    sigma, ycbcr, wavelet, method, mode, shifts = params
    for channel in range(3):
      self.widgets.entries[channel].set_name("")
      self.widgets.entries[channel].set_text_block(f"{sigma[channel]:.5e}")
    if sigma[1] != sigma[0] or sigma[2] != sigma[0]: self.widgets.bindbutton.set_active_block(False)
    self.widgets.ycbcrbutton.set_active(ycbcr)
    self.widgets.waveletcombo.set_selected(wavelet)
    self.widgets.methodbuttons.set_selected(method)
    self.widgets.modebuttons.set_selected(mode)
    self.widgets.shiftsbutton.set_value(shifts)

  def run(self, params):
    """Run tool for parameters 'params'."""
    sigma, ycbcr, wavelet, method, mode, shifts = params
    kwargs = dict(channel_axis = -1, sigma = sigma, wavelet = wavelet, mode = mode.lower(), wavelet_levels = None,
                  convert2ycbcr = ycbcr, method = method, rescale_sigma = True)
    self.image.rgb = cycle_spin(self.reference.rgb, channel_axis = 0, max_shifts = shifts, func = denoise_wavelet, func_kw = kwargs, num_workers = None)
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    sigma, ycbcr, wavelet, method, mode, shifts = params
    operation = f"WaveletsFilter(R = {sigma[0]:.5e}, G = {sigma[1]:.5e}, B = {sigma[2]:.5e}, "
    if ycbcr: operation += " YCbCr,"
    operation += f" wavelets = {wavelet}, method = {method}, mode = {mode}, shifts = {shifts})"
    return operation

 # Update widgets.

  def update(self, changed):
    """Update widgets on change of 'changed'."""
    text = self.widgets.entries[changed].get_text()
    try:
      value = float(text)
    except:
      self.widgets.entries[changed].set_name("red-entry")
      return
    self.widgets.entries[changed].set_name("")
    if self.widgets.bindbutton.get_active():
      for channel in range(3):
        self.widgets.entries[channel].set_text_block(text)
